#  ============================================================================
#
#  Copyright (C) 2007-2011 Conceptive Engineering bvba. All rights reserved.
#  www.conceptive.be / project-camelot@conceptive.be
#
#  This file is part of the Camelot Library.
#
#  This file may be used under the terms of the GNU General Public
#  License version 2.0 as published by the Free Software Foundation
#  and appearing in the file license.txt included in the packaging of
#  this file.  Please review this information to ensure GNU
#  General Public Licensing requirements will be met.
#
#  If you are unsure which license is appropriate for your use, please
#  visit www.python-camelot.com or contact project-camelot@conceptive.be
#
#  This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
#  WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
#
#  For use of this library in commercial applications, please contact
#  project-camelot@conceptive.be
#
#  ============================================================================

"""Convenience functions and classes to present views to the user"""

from PyQt4 import QtGui
from PyQt4 import QtCore
from PyQt4.QtCore import Qt

import logging
logger = logging.getLogger('camelot.view.workspace')

from camelot.admin.action import ApplicationActionGuiContext
from camelot.core.utils import ugettext as _
from camelot.view.model_thread import gui_function, post
from camelot.view.controls.action_widget import ( ActionLabel, 
                                                  HOVER_ANIMATION_DISTANCE,
                                                  NOTIFICATION_ANIMATION_DISTANCE )

from camelot.view.art import Icon

class DesktopBackground(QtGui.QWidget):
    """
    A custom background widget for the desktop. This widget is contained
    by the first tab ('Start' tab) of the desktop workspace.
    """
    
    def __init__( self, gui_context, parent ):
        super(DesktopBackground, self).__init__( parent )
        self.gui_context = gui_context
        mainLayout = QtGui.QVBoxLayout()
        
        actionButtonsLayout = QtGui.QGridLayout()
        actionButtonsLayout.setObjectName('actionButtonsLayout')
        actionButtonsLayout.setContentsMargins(200, 20, 200, 20)
        
        actionButtonInfoWidget = ActionButtonInfoWidget()
        actionButtonInfoWidget.setObjectName('actionButtonInfoWidget')
                
        mainLayout.addWidget(actionButtonInfoWidget, 0, Qt.AlignCenter)
        mainLayout.addLayout(actionButtonsLayout)
        
        self.setLayout(mainLayout)
        
        # Set a white background color
        palette = self.palette()
        self.setAutoFillBackground(True)
        palette.setBrush(QtGui.QPalette.Window, Qt.white)
        self.setPalette(palette)
        
    # This method is invoked when the desktop workspace decides or gets told
    # that the actions should be updated due to the presence of to be added 
    # actions. 
    @QtCore.pyqtSlot(object)
    def set_actions(self, actions):
        """
        :param actions: a list of EntityActions
        """
        #
        # Remove old actions
        #
        for actionButton in self.findChildren(ActionLabel):
            actionButton.deleteLater()
        
        # Make sure that the action buttons aren't visually split
        # up in two rows when there are e.g. only 3 of them.
        # So:
        #  <= 3 action buttons: 1 row and 1, 2 or 3 columns;
        #  >= 4 action buttons: 2 rows and 2, 3, 4 or 5 columns.
        actionButtonsLayoutMaxItemsPerRowCount = max((len(actions) + 1) / 2, 3)
        
        actionButtonsLayout = self.findChild(QtGui.QGridLayout, 'actionButtonsLayout')
        if actionButtonsLayout is not None:
            for position in xrange(0, min( len(actions), actionButtonsLayoutMaxItemsPerRowCount) ):
                action = actions[position]
                actionButton = action.render( self.gui_context, self )
                actionButton.entered.connect(self.onActionButtonEntered)
                actionButton.left.connect(self.onActionButtonLeft)
                actionButton.setInteractive(True)
                actionButtonsLayout.addWidget(ActionButtonContainer(actionButton), 0, position, Qt.AlignCenter)

            for position in xrange(actionButtonsLayoutMaxItemsPerRowCount, len(actions)):
                action = actions[position]
                actionButton = action.render( self.gui_context, self )
                actionButton.entered.connect(self.onActionButtonEntered)
                actionButton.left.connect(self.onActionButtonLeft)
                actionButton.setInteractive(True)
                actionButtonsLayout.addWidget(ActionButtonContainer(actionButton), 1, position % actionButtonsLayoutMaxItemsPerRowCount, Qt.AlignCenter)
            
    @QtCore.pyqtSlot()
    def onActionButtonEntered(self):
        actionButton = self.sender()
        actionButtonInfoWidget = self.findChild(QtGui.QWidget, 'actionButtonInfoWidget')
        if actionButtonInfoWidget is not None:
            # @todo : get state should be called with a model context as first
            #         argument
            post( actionButton.action.get_state,
                  actionButtonInfoWidget.setInfoFromState,
                  args = (None,) )
       
    @QtCore.pyqtSlot()
    def onActionButtonLeft(self):
        actionButtonInfoWidget = self.findChild(QtGui.QWidget, 'actionButtonInfoWidget')
        if actionButtonInfoWidget is not None:
            actionButtonInfoWidget.resetInfo()
        
    # This custom event handler makes sure that the action buttons aren't
    # drawn in the wrong position on this widget after the screen has been
    # e.g. maximized or resized by using the window handles.
    
    def resizeEvent(self, event):
        for actionButton in self.findChildren(ActionLabel):
            actionButton.resetLayout()

        event.ignore()

    # This slot is called after the navpane's animation has finished. During 
    # this sliding animation, all action buttons are linearly moved to the right,
    # giving the user a small window in which he or she may cause visual problems
    # by already hovering the action buttons. This switch assures that the user 
    # cannot perform mouse interaction with the action buttons until they're
    # static.
    @QtCore.pyqtSlot()
    def makeInteractive(self, interactive=True):
        for actionButton in self.findChildren(ActionLabel):
            actionButton.setInteractive(interactive)
            
    def refresh(self):
        pass
            
class ActionButtonContainer(QtGui.QWidget):
    def __init__(self, actionButton, parent = None):
        super(ActionButtonContainer, self).__init__(parent)
        
        mainLayout = QtGui.QHBoxLayout()
        # Set some margins to avoid the ActionButton being visually clipped
        # when performing the hoverAnimation.
        mainLayout.setContentsMargins(2*NOTIFICATION_ANIMATION_DISTANCE,
                                      2*HOVER_ANIMATION_DISTANCE,
                                      2*NOTIFICATION_ANIMATION_DISTANCE,
                                      2*HOVER_ANIMATION_DISTANCE)
        mainLayout.addWidget(actionButton)
        self.setLayout(mainLayout)
        
    def mousePressEvent(self, event):
        # Send this event to the ActionButton that is contained by this widget.
        self.layout().itemAt(0).widget().onContainerMousePressEvent(event)
            
class ActionButtonInfoWidget(QtGui.QWidget):
    def __init__(self, parent = None):
        super(ActionButtonInfoWidget, self).__init__(parent)
        
        mainLayout = QtGui.QHBoxLayout()
        
        font = self.font()
        font.setPointSize(14)
        
        actionNameLabel = QtGui.QLabel()
        actionNameLabel.setFont(font)
        actionNameLabel.setFixedSize(250, 50)
        actionNameLabel.setAlignment(Qt.AlignCenter)
        actionNameLabel.setObjectName('actionNameLabel')
        
        actionDescriptionLabel = QtGui.QLabel()
        actionDescriptionLabel.setFixedSize(250, 200)
        actionDescriptionLabel.setObjectName('actionDescriptionLabel')

        mainLayout.addWidget(actionNameLabel, 0, Qt.AlignVCenter)
        mainLayout.addWidget(actionDescriptionLabel)

        self.setLayout(mainLayout)

    @QtCore.pyqtSlot( object )
    def setInfoFromState(self, state):
        actionNameLabel = self.findChild(QtGui.QLabel, 'actionNameLabel')
        if actionNameLabel is not None:
            actionNameLabel.setText( unicode( state.verbose_name ) )
        
        actionDescriptionLabel = self.findChild(QtGui.QLabel, 'actionDescriptionLabel')
        if actionDescriptionLabel is not None:
            tooltip = unicode( state.tooltip or '' )
            actionDescriptionLabel.setText(tooltip)
            if tooltip:
                # Do not use show() or hide() in this case, since it will
                # cause the actionButtons to be drawn on the wrong position.
                # Instead, just set the width of the widget to either 0 or 250.
                actionDescriptionLabel.setFixedWidth(250)
            else:
                actionDescriptionLabel.setFixedWidth(0)
            
    def resetInfo(self):
        actionNameLabel = self.findChild(QtGui.QLabel, 'actionNameLabel')
        if actionNameLabel is not None:
            actionNameLabel.setText('')
        
        actionDescriptionLabel = self.findChild(QtGui.QLabel, 'actionDescriptionLabel')
        if actionDescriptionLabel is not None:
            actionDescriptionLabel.setText('')
    
class DesktopTabbar(QtGui.QTabBar):

    change_view_mode_signal = QtCore.pyqtSignal()
    
    def mouseDoubleClickEvent(self, event):
        self.change_view_mode_signal.emit()
        event.accept()
        
    def tabSizeHint(self, index):
        originalSizeHint = super(DesktopTabbar, self).tabSizeHint(index)
        minimumWidth = max(160, originalSizeHint.width())
        
        return QtCore.QSize(minimumWidth, originalSizeHint.height())

class DesktopWorkspace(QtGui.QWidget):
    """
    A tab based workspace that can be used by views to display themselves.
    
    In essence this is a wrapper around QTabWidget to do some initial setup
    and provide it with a background widget.
    This was originallly implemented using the QMdiArea, but the QMdiArea has 
    too many drawbacks, like not being able to add close buttons to the tabs
    in a decent way.

    .. attribute:: background

    The widget class to be used as the view for the uncloseable 'Start' tab.
    """

    view_activated_signal = QtCore.pyqtSignal(QtGui.QWidget)
    change_view_mode_signal = QtCore.pyqtSignal()
    last_view_closed_signal = QtCore.pyqtSignal()

    @gui_function
    def __init__(self, app_admin, parent):
        super(DesktopWorkspace, self).__init__(parent)
        self.gui_context = ApplicationActionGuiContext()
        self.gui_context.admin = app_admin
        self.gui_context.desktop = self
        self._app_admin = app_admin
        
        layout = QtGui.QHBoxLayout()
        layout.setContentsMargins(0, 0, 0, 0)
        layout.setSpacing(0)
        
        # Setup the tab widget
        self._tab_widget = QtGui.QTabWidget( self )
        tab_bar = DesktopTabbar(self._tab_widget)
        tab_bar.setToolTip(_('Double click to (un)maximize'))
        tab_bar.change_view_mode_signal.connect(self._change_view_mode)
        self._tab_widget.setTabBar(tab_bar)
        self._tab_widget.setDocumentMode(True)
        self._tab_widget.setTabsClosable(True)
        self._tab_widget.tabCloseRequested.connect(self._tab_close_request)
        self._tab_widget.currentChanged.connect(self._tab_changed)
        layout.addWidget(self._tab_widget)
        
        # Setup the background widget
        self._background_widget = DesktopBackground( self.gui_context, self )
        self._app_admin.actions_changed_signal.connect(self.reload_background_widget)
        self._tab_widget.addTab(self._background_widget,
                                Icon('tango/16x16/actions/go-home.png').getQIcon(),
                                _('Home'))
        if tab_bar.tabButton(0, QtGui.QTabBar.RightSide):
            tab_bar.tabButton(0, QtGui.QTabBar.RightSide).hide()
        elif tab_bar.tabButton(0, QtGui.QTabBar.LeftSide):
            # mac for example has the close button on the left side by default
            tab_bar.tabButton(0, QtGui.QTabBar.LeftSide).hide()
        
        self.setLayout(layout)
        self.reload_background_widget()
             
    @QtCore.pyqtSlot()
    def reload_background_widget(self):
        post(self._app_admin.get_actions, self._background_widget.set_actions)

    @QtCore.pyqtSlot()
    def _change_view_mode(self):
        self.change_view_mode_signal.emit()
        
    @QtCore.pyqtSlot(int)
    def _tab_close_request(self, index):
        """
        Handle the request for the removal of a tab at index.
        
        Note that only at-runtime added tabs are being closed, implying
        the immortality of the 'Start' tab.
        """
        if index > 0:
            view = self._tab_widget.widget(index)
            if view:
                # it's not enough to simply remove the tab, because this
                # would keep the underlying view widget alive
                view.deleteLater()
                self._tab_widget.removeTab(index)

    @QtCore.pyqtSlot(int)
    def _tab_changed(self, _index):
        """
        The active tab has changed, emit the view_activated signal.
        """
        self.view_activated_signal.emit(self.active_view())

    def active_view(self):
        """
        :return: The currently active view or None in case of the 'Start' tab.
        """
        i = self._tab_widget.currentIndex()
        
        if i == 0: # 'Start' tab
            return None
        
        return self._tab_widget.widget(i)

    @QtCore.pyqtSlot(QtCore.QString)
    def change_title(self, new_title):
        """
        Slot to be called when the tile of a view needs to change.
        
        Note: the title of the 'Start' tab cannot be overwritten.
        """
        # the request of the sender does not work in older pyqt versions
        # therefore, take the current index, notice this is not correct !!
        #
        # sender = self.sender()
        sender = self.active_view()
        
        if sender:
            index = self._tab_widget.indexOf(sender)
            if index > 0:
                self._tab_widget.setTabText(index, new_title)
                
    @QtCore.pyqtSlot(QtGui.QIcon)
    def change_icon(self, new_icon):
        """
        Slot to be called when the icon of a view needs to change.
        
        Note: the icon of the 'Start' tab cannot be overwritten.
        """
        # the request of the sender does not work in older pyqt versions
        # therefore, take the current index, notice this is not correct !!
        #
        # sender = self.sender()
        sender = self.active_view()
        
        if sender:
            index = self._tab_widget.indexOf(sender)
            if index > 0:
                self._tab_widget.setTabIcon(index, new_icon)

    def set_view(self, view, icon = None, title = '...'):
        """
        Remove the currently active view and replace it with a new view.
        """
        index = self._tab_widget.currentIndex()
        
        if index == 0: # 'Start' tab is currently visible.
            self.add_view(view, icon, title)
        else:
            self._tab_close_request(index)
            view.title_changed_signal.connect(self.change_title)
            view.icon_changed_signal.connect(self.change_icon)
            if icon:
                index = self._tab_widget.insertTab(index, view, icon, title)
            else:
                index = self._tab_widget.insertTab(index, view, title)
            self._tab_widget.setCurrentIndex(index)

    @gui_function
    def add_view(self, view, icon = None, title = '...'):
        """
        Add a Widget implementing AbstractView to the workspace.
        """
        view.title_changed_signal.connect(self.change_title)
        view.icon_changed_signal.connect(self.change_icon)
        if icon:
            index = self._tab_widget.addTab(view, icon, title)
        else:
            index = self._tab_widget.addTab(view, title)
        self._tab_widget.setCurrentIndex(index)

    def refresh(self):
        """Refresh all views on the desktop"""
        for i in range( self._tab_widget.count() ):
            self._tab_widget.widget(i).refresh()
            
    def close_all_views(self):
        """
        Remove all views, except the 'Start' tab, from the workspace.
        """
        # NOTE: will call removeTab until tab widget is cleared
        # but removeTab does not really delete the page objects
        #self._tab_widget.clear()
        max_index = self._tab_widget.count()
        
        while max_index > 0:
            self._tab_widget.tabCloseRequested.emit(max_index)
            max_index -= 1

top_level_windows = []

def show_top_level(view, parent):
    """Show a widget as a top level window
    :param view: the widget extend AbstractView
    :param parent: the widget with regard to which the top level
    window will be placed.
     """
    from camelot.view.register import register
    #
    # Register the view with reference to itself.  This will keep
    # the Python object alive as long as the Qt object is not
    # destroyed.  Hence Python will not trigger the deletion of the
    # view as long as the window is not closed
    #
    register( view, view )
    #
    # set the parent to None to avoid the window being destructed
    # once the parent gets destructed
    #
    view.setParent( None )
    view.setWindowFlags(QtCore.Qt.Window)
    #
    # Make the window title blank to prevent the something
    # like main.py or pythonw being displayed
    #
    view.setWindowTitle( u'' )
    view.title_changed_signal.connect( view.setWindowTitle )
    view.icon_changed_signal.connect( view.setWindowIcon )
    view.setAttribute(QtCore.Qt.WA_DeleteOnClose)
    #
    # position the new window in the center of the same screen
    # as the parent
    #
    screen = QtGui.QApplication.desktop().screenNumber(parent)
    available = QtGui.QApplication.desktop().availableGeometry(screen)

    point = QtCore.QPoint(available.x() + available.width()/2,
                          available.y() + available.height()/2)
    point = QtCore.QPoint(point.x()-view.width()/2,
                          point.y()-view.height()/2)
    view.move( point )

    #view.setWindowModality(QtCore.Qt.WindowModal)
    view.show()



from camelot.admin.object_admin import ObjectAdmin
from camelot.view.controls import delegates

class Coordinate(object):
  
  def __init__(self):
    self.id = 1
    self.x = 0.0
    self.y = 0.0
    
  def _get_x(self):
    return self.x
  
  def _set_x(self, x):
    self.x = x
    self.y = max(self.y,x)
    
  _x = property(_get_x, _set_x)
      
  class Admin(ObjectAdmin):
    form_display = ['_x', 'y',]
    field_attributes = dict(_x=dict(delegate=delegates.FloatDelegate, name='x'),
                            y=dict(delegate=delegates.FloatDelegate),)
    form_size = (100,100)



#  ============================================================================
#
#  Copyright (C) 2007-2011 Conceptive Engineering bvba. All rights reserved.
#  www.conceptive.be / project-camelot@conceptive.be
#
#  This file is part of the Camelot Library.
#
#  This file may be used under the terms of the GNU General Public
#  License version 2.0 as published by the Free Software Foundation
#  and appearing in the file license.txt included in the packaging of
#  this file.  Please review this information to ensure GNU
#  General Public Licensing requirements will be met.
#
#  If you are unsure which license is appropriate for your use, please
#  visit www.python-camelot.com or contact project-camelot@conceptive.be
#
#  This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
#  WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
#
#  For use of this library in commercial applications, please contact
#  project-camelot@conceptive.be
#
#  ============================================================================


import logging

from PyQt4 import QtGui, QtCore
from PyQt4.QtGui import QDialog
from PyQt4.QtGui import QBoxLayout
from PyQt4.QtGui import QVBoxLayout
from PyQt4.QtGui import QPushButton

logger = logging.getLogger('camelot.view.controls.combobox_input_dialog')

class ComboBoxInputDialog(QDialog):

    def __init__(self, autoaccept=False, parent=None):
        """
        :param autoaccept: if True, the value of the ComboBox is immediately
        accepted after selecting it.
        """
        super(ComboBoxInputDialog, self).__init__(parent)
        self._autoaccept = autoaccept
        layout = QVBoxLayout()
        label = QtGui.QLabel()
        label.setObjectName( 'label' )
        combobox = QtGui.QComboBox()
        combobox.setObjectName( 'combobox' )
        combobox.activated.connect( self._combobox_activated )
        ok_button = QPushButton('OK')
        ok_button.setObjectName( 'ok' )
        cancel_button = QPushButton('Cancel')
        cancel_button.setObjectName( 'cancel' )
        ok_button.pressed.connect(self.accept)
        cancel_button.pressed.connect(self.reject)
        button_layout = QBoxLayout(QBoxLayout.RightToLeft)
        button_layout.addWidget(cancel_button)
        button_layout.addWidget(ok_button)
        layout.addWidget( label )
        layout.addWidget( combobox )
        layout.addLayout( button_layout )
        self.setLayout( layout )

    @QtCore.pyqtSlot(int)
    def _combobox_activated(self, index):
        if self._autoaccept:
            self.accept()
        print index
        
    def set_label_text(self, text):
        label = self.findChild( QtGui.QWidget, 'label' )
        if label != None:
            label.setText( text )

    def set_items(self, items):
        combobox = self.findChild( QtGui.QWidget, 'combobox' )
        if combobox != None:
            combobox.addItems(items)

    def count(self):
        combobox = self.findChild( QtGui.QWidget, 'combobox' )
        if combobox != None:
            return combobox.count()
        return 0

    def set_data(self, index, data, role):
        combobox = self.findChild( QtGui.QWidget, 'combobox' )
        if combobox != None:
            combobox_model = combobox.model()
            model_index = combobox_model.index(index, 0)
            combobox_model.setData(model_index, data, role)

    def get_text(self):
        combobox = self.findChild( QtGui.QWidget, 'combobox' )
        if combobox != None:
            return combobox.currentText()

    def set_ok_button_default(self):
        ok = self.findChild( QtGui.QWidget, 'ok' )
        if ok != None:
            ok.setFocus()

    def set_cancel_button_default(self):
        cancel = self.findChild( QtGui.QWidget, 'cancel' )
        if cancel != None:
            cancel.setFocus()

    def set_ok_button_text(self, text):
        ok = self.findChild( QtGui.QWidget, 'ok' )
        if ok != None:
            ok.setText(text)

    def set_cancel_button_text(self, text):
        cancel = self.findChild( QtGui.QWidget, 'cancel' )
        if cancel != None:
            cancel.setText(text)

    def set_window_title(self, title):
        self.setWindowTitle(title)
    
    def set_choice_by_text(self, text):
        combobox = self.findChild( QtGui.QWidget, 'combobox' )
        if combobox != None:
            index = combobox.findText(text)
            self.set_choice_by_index(index)
    
    def set_choice_by_index(self, index):
        combobox = self.findChild( QtGui.QWidget, 'combobox' )
        if combobox != None:
            combobox.setCurrentIndex(index)
        else:
            raise Exception('No combobox to set the choice')


#  ============================================================================
#
#  Copyright (C) 2007-2011 Conceptive Engineering bvba. All rights reserved.
#  www.conceptive.be / project-camelot@conceptive.be
#
#  This file is part of the Camelot Library.
#
#  This file may be used under the terms of the GNU General Public
#  License version 2.0 as published by the Free Software Foundation
#  and appearing in the file license.txt included in the packaging of
#  this file.  Please review this information to ensure GNU
#  General Public Licensing requirements will be met.
#
#  If you are unsure which license is appropriate for your use, please
#  visit www.python-camelot.com or contact project-camelot@conceptive.be
#
#  This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
#  WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
#
#  For use of this library in commercial applications, please contact
#  project-camelot@conceptive.be
#
#  ============================================================================

from camelot.admin.action.base import Action
from application_action import ( ApplicationActionGuiContext,
                                 ApplicationActionModelContext )

class FormActionModelContext( ApplicationActionModelContext ):
    """On top of the attributes of the 
    :class:`camelot.admin.action.application_action.ApplicationActionModelContext`, 
    this context contains :
        
    .. attribute:: current_row
    
        the row in the list that is currently displayed in the form
        
    .. attribute:: collection_count
    
        the number of objects that can be reached in the form.
        
    .. attribute:: selection_count
    
        the number of objects displayed in the form, at most 1.
                
    .. attribute:: session
    
        The session to which the objects in the list belong.
        
    The :attr:`selection_count` attribute allows the 
    :meth:`model_run` to quickly evaluate the size of the collection without 
    calling the potetially time consuming method :meth:`get_collection`.
    """
    
    def __init__( self ):
        super( FormActionModelContext, self ).__init__()
        self._model = None
        self.admin = None
        self.current_row = None
        self.collection_count = 0
        self.selection_count = 1
        
    @property
    def session( self ):
        return self._model.admin.get_query().session
        
    def get_object( self ):
        """
        :return: the object currently displayed in the form, None if no object
            is displayed yet
        """
        if self.current_row != None:
            return self._model._get_object( self.current_row )
    
    def get_collection( self, yield_per = None ):
        """
        :param yield_per: an integer number giving a hint on how many objects
            should fetched from the database at the same time.
        :return: a generator over the objects in the list
        """
        for obj in self._model.get_collection():
            yield obj
            
    def get_selection( self, yield_per = None ):
        """
        Method to be compatible with a 
        :class:`camelot.admin.action.list_action.ListActionModelContext`, this
        allows creating a single Action to be used on a form and on list.
        
        :param yield_per: this parameter has no effect, it's here only for
            compatibility with :meth:`camelot.admin.action.list_action.ListActionModelContext.get_selection`
        :return: a generator that yields the current object displayed in the 
            form and does not yield anything if no object is displayed yet
            in the form.
        """
        if self.current_row != None:
            yield self._model._get_object( self.current_row )
        
class FormActionGuiContext( ApplicationActionGuiContext ):
    """The context for an :class:`Action` on a form.  On top of the attributes of the 
    :class:`camelot.admin.action.application_action.ApplicationActionGuiContext`, 
    this context contains :

    .. attribute:: widget_mapper

       the :class:`QtGui.QDataWidgetMapper` class that relates the form 
       widget to the model.
       
    """
        
    model_context = FormActionModelContext
    
    def __init__( self ):
        super( FormActionGuiContext, self ).__init__()
        self.widget_mapper = None

    def create_model_context( self ):
        context = super( FormActionGuiContext, self ).create_model_context()
        context._model = self.widget_mapper.model()
        context.collection_count = context._model.rowCount()
        context.current_row = self.widget_mapper.currentIndex()
        return context
        
    def copy( self ):
        new_context = super( FormActionGuiContext, self ).copy()
        new_context.widget_mapper = self.widget_mapper
        return new_context

def structure_to_form_actions( structure ):
    """Convert a list of python objects to a list of form actions.  If the python
    object is a tuple, a CallMethod is constructed with this tuple as arguments.  If
    the python object is an instance of as Action, it is kept as is.
    """
    from list_action import CallMethod
    
    def object_to_action( o ):
        if isinstance( o, Action ):
            return o
        return CallMethod( o[0], o[1] )

    return [object_to_action( o ) for o in structure]



#  ============================================================================
#
#  Copyright (C) 2007-2011 Conceptive Engineering bvba. All rights reserved.
#  www.conceptive.be / project-camelot@conceptive.be
#
#  This file is part of the Camelot Library.
#
#  This file may be used under the terms of the GNU General Public
#  License version 2.0 as published by the Free Software Foundation
#  and appearing in the file license.txt included in the packaging of
#  this file.  Please review this information to ensure GNU
#  General Public Licensing requirements will be met.
#
#  If you are unsure which license is appropriate for your use, please
#  visit www.python-camelot.com or contact project-camelot@conceptive.be
#
#  This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
#  WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
#
#  For use of this library in commercial applications, please contact
#  project-camelot@conceptive.be
#
#  ============================================================================

from customdelegate import CustomDelegate, DocumentationMetaclass
from PyQt4.QtCore import Qt

from camelot.view.controls import editors
from camelot.core.utils import variant_to_pyobject
from camelot.view.proxy import ValueLoading

class EnumerationDelegate(CustomDelegate):
    """Contrary to the comboboxdelegate, the enumeration delegate does not support dynamic
    choices"""

    __metaclass__ = DocumentationMetaclass
    editor = editors.ChoicesEditor

    def __init__(self, parent=None, choices=[], editable=True, **kwargs):
        CustomDelegate.__init__(self, parent, editable=editable, **kwargs)
        self.choices = choices
        self._choices_dict = dict(choices)
        
    def createEditor(self, parent, option, index):
        editor = super(EnumerationDelegate, self).createEditor(parent, option, index)
        editor.set_choices(self.choices)
        return editor

    def paint(self, painter, option, index):
        value = variant_to_pyobject(index.data(Qt.EditRole))
        if value==ValueLoading:
            value = ''
        else:
            value = unicode(self._choices_dict.get(value, '...'))
        painter.save()
        self.paint_text(painter, option, index, value)
        painter.restore()





#  ============================================================================
#
#  Copyright (C) 2007-2011 Conceptive Engineering bvba. All rights reserved.
#  www.conceptive.be / project-camelot@conceptive.be
#
#  This file is part of the Camelot Library.
#
#  This file may be used under the terms of the GNU General Public
#  License version 2.0 as published by the Free Software Foundation
#  and appearing in the file license.txt included in the packaging of
#  this file.  Please review this information to ensure GNU
#  General Public Licensing requirements will be met.
#
#  If you are unsure which license is appropriate for your use, please
#  visit www.python-camelot.com or contact project-camelot@conceptive.be
#
#  This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
#  WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
#
#  For use of this library in commercial applications, please contact
#  project-camelot@conceptive.be
#
#  ============================================================================

from functools import wraps

from PyQt4 import QtGui
from PyQt4 import QtCore

import logging
logger = logging.getLogger('camelot.view.model_thread')

_model_thread_ = []

# this might be set to False, for unittesting purpose
verify_threads = True

class ModelThreadException(Exception):
    pass


def model_function(original_function):
    """Decorator to ensure a function is only called from within the model
    thread. If this function is called in another thread, an exception will be
    thrown"""

    def in_model_thread():
        """return wether current thread is model thread"""
        from no_thread_model_thread import NoThreadModelThread
        current_thread = QtCore.QThread.currentThread()
        model_thread = get_model_thread()
        return (current_thread==model_thread) or isinstance(
            model_thread, (NoThreadModelThread,)
        )

    @wraps(original_function)
    def wrapper(*args, **kwargs):
        assert (not verify_threads) or in_model_thread()
        return original_function(*args, **kwargs)

    return wrapper


def gui_function(original_function):
    """Decorator to ensure a function is only called from within the gui
    thread. If this function is called in another thread, an exception will be
    thrown"""

    def in_gui_thread():
        """return wether current thread is gui thread"""
        gui_thread = QtGui.QApplication.instance().thread()
        current_thread = QtCore.QThread.currentThread()
        return gui_thread == current_thread

    @wraps(original_function)
    def wrapper(*args, **kwargs):
        assert (not verify_threads) or in_gui_thread()
        return original_function(*args, **kwargs)

    return wrapper


def setup_model():
    """Call the setup_model function in the settings"""
    from camelot.core.conf import settings
    settings.setup_model()

class AbstractModelThread(QtCore.QThread):
    """Abstract implementation of a model thread class
    Thread in which the model runs, all requests to the model should be
    posted to the the model thread.

    This class ensures the gui thread doesn't block when the model needs
    time to complete tasks by providing asynchronous communication between
    the model thread and the gui thread
    
    The Model thread class provides a number of signals :
    
    *thread_busy_signal*
    
    indicates if the model thread is working in the background
    
    *setup_exception_signal*
    
    this signal is emitted when there was an exception setting up the model
    thread, eg no connection to the database could be made.  this exception
    is mostly fatal for the application.
    """

    thread_busy_signal = QtCore.pyqtSignal(bool)
    setup_exception_signal = QtCore.pyqtSignal(object)

    def __init__(self, setup_thread=setup_model):
        """:param setup_thread: function to be called at startup of the thread
        to initialize everything, by default this will setup the model. Set to
        None if nothing should be done."""
        super(AbstractModelThread, self).__init__()
        self.logger = logging.getLogger(logger.name + '.%s' % id(self))
        self._setup_thread = setup_thread
        self._exit = False
        self._traceback = ''
        self.logger.debug('model thread constructed')

    def run(self):
        pass

    def traceback(self):
        """The formatted traceback of the last exception in the model thread"""
        return self._traceback

    def wait_on_work(self):
        """Wait for all work to be finished, this function should only be used
    to do unit testing and such, since it will block the calling thread until
    all work is done"""
        pass

    def post(self, request, response=None, exception=None, args=()):
        """Post a request to the model thread, request should be a function
        that takes no arguments. The request function will be called within the
        model thread. When the request is finished, on first occasion, the
        response function will be called within the gui thread. The response
        function takes as arguments, the results of the request function.

        :param request: function to be called within the model thread
        :param response: a slot that will be called with the result of the
        request function
        :param exception: a slot that will be called in case request throws an
        exception
        :param args: arguments with which the request function will be called        
        """
        raise NotImplemented

    def busy(self):
        """Return True or False indicating wether either the model or the gui
        thread is doing something"""
        return False
    
    def stop(self):
        """Stop the model thread from accepting any further posts.
        """
        return True

def construct_model_thread(*args, **kwargs):
    from signal_slot_model_thread import SignalSlotModelThread
    _model_thread_.insert(0, SignalSlotModelThread(*args, **kwargs))

def has_model_thread():
    return len(_model_thread_) > 0

def get_model_thread():
    return _model_thread_[0]

def post(request, response=None, exception=None, args=()):
    """Post a request and a response to the default model thread"""
    mt = get_model_thread()
    mt.post(request, response, exception, args)



#  ============================================================================
#
#  Copyright (C) 2007-2011 Conceptive Engineering bvba. All rights reserved.
#  www.conceptive.be / project-camelot@conceptive.be
#
#  This file is part of the Camelot Library.
#
#  This file may be used under the terms of the GNU General Public
#  License version 2.0 as published by the Free Software Foundation
#  and appearing in the file license.txt included in the packaging of
#  this file.  Please review this information to ensure GNU
#  General Public Licensing requirements will be met.
#
#  If you are unsure which license is appropriate for your use, please
#  visit www.python-camelot.com or contact project-camelot@conceptive.be
#
#  This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
#  WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
#
#  For use of this library in commercial applications, please contact
#  project-camelot@conceptive.be
#
#  ============================================================================

"""The action module contains various QAction classes, representing commands
that can be invoked via menus, toolbar buttons, and keyboard shortcuts."""

from PyQt4 import QtGui
from PyQt4.QtCore import Qt

from camelot.view.art import Icon
from camelot.core.utils import ugettext as _

class ActionFactory(object):
    """Utility class to generate some default actions we need
    in several places.
    
    Each method of this class, returns a certain action with
    a default text, icon and shortcut.
    """

    icon_backup = Icon('tango/16x16/actions/document-save.png')
    icon_restore = Icon('tango/16x16/devices/drive-harddisk.png')
    icon_pgsetup = Icon('tango/16x16/actions/document-properties.png')
    icon_print = Icon('tango/16x16/actions/document-print.png')
    icon_preview = Icon('tango/16x16/actions/document-print-preview.png')
    icon_copy = Icon('tango/16x16/actions/edit-copy.png')
    icon_new = Icon('tango/16x16/actions/document-new.png')
    icon_delete = Icon('tango/16x16/places/user-trash.png')
    icon_excel = Icon('tango/16x16/mimetypes/x-office-spreadsheet.png')
    icon_word = Icon('tango/16x16/mimetypes/x-office-document.png')
    icon_mail = Icon('tango/16x16/actions/mail-message-new.png')
    icon_import = Icon('tango/16x16/mimetypes/text-x-generic.png')
    icon_help = Icon('tango/16x16/apps/help-browser.png')
        
    @classmethod
    def create_action(*a, **kw):
        """creates and returns a QAction object"""
        # collect params
        parent = kw['parent']
        text = kw['text']
        slot = kw.get('slot', None)
        shortcut = kw.get('shortcut', '')
        actionicon = kw.get('actionicon', '')
        tip = kw.get('tip', '')
        checkable = kw.get('checkable', False)
        #signal = kw.get('signal', 'triggered()')
        widgetaction = kw.get('widgetaction', False)
        if widgetaction:
            action = QtGui.QWidgetAction(parent)
        else:
            action = QtGui.QAction(parent)
        action.setText(text)
        if actionicon:
            action.setIcon(actionicon.getQIcon())
        if shortcut:
            action.setShortcut(shortcut)
        if tip:
            action.setToolTip(tip)
            action.setStatusTip(tip)
        if slot is not None:
            action.triggered.connect( slot )
        if checkable:
            action.setCheckable(True)
        return action

    @classmethod
    def copy(cls, parent, slot, **kwargs):
        default = dict(
            text=_('Copy'),
            slot=slot,
            parent=parent,
            shortcut=QtGui.QKeySequence.Copy,
            actionicon=Icon('tango/16x16/actions/edit-copy.png'),
            tip=_('Duplicate')
        )
        default.update(kwargs)
        return cls.create_action(**default)

    @classmethod
    def paste(cls, parent, slot, **kwargs):
        default = dict(
            text=_('Paste'),
            slot=slot,
            parent=parent,
            shortcut=QtGui.QKeySequence.Paste,
            actionicon=Icon('tango/16x16/actions/edit-paste.png'),
            tip=_('Paste content from clipboard')
        )
        default.update(kwargs)
        return cls.create_action(**default)
            
    @classmethod
    def view_first(cls, parent, slot, **kwargs):
        default = dict(
            text=_('First'),
            slot=slot,
            parent=parent,
            shortcut=QtGui.QKeySequence.MoveToStartOfDocument,
            actionicon=Icon('tango/16x16/actions/go-first.png'),
            tip=_('First')
        )
        default.update(kwargs)
        return cls.create_action(**default)

    @classmethod
    def view_last(cls, parent, slot, **kwargs):
        default = dict(
            text=_('Last'),
            slot=slot,
            parent=parent,
            shortcut=QtGui.QKeySequence.MoveToEndOfDocument,
            actionicon=Icon('tango/16x16/actions/go-last.png'),
            tip=_('Last')
        )
        default.update(kwargs)
        return cls.create_action(**default)

    @classmethod
    def view_next(cls, parent, slot, **kwargs):
        default = dict(
            text=_('Next'),
            slot=slot,
            parent=parent,
            shortcut=QtGui.QKeySequence.MoveToNextPage,
            actionicon=Icon('tango/16x16/actions/go-next.png'),
            tip=_('Next')
        )
        default.update(kwargs)
        return cls.create_action(**default)

    @classmethod
    def view_previous(cls, parent, slot, **kwargs):
        default = dict(
            text=_('Previous'),
            slot=slot,
            parent=parent,
            shortcut=QtGui.QKeySequence.MoveToPreviousPage,
            actionicon=Icon('tango/16x16/actions/go-previous.png'),
            tip=_('Previous')
        )
        default.update(kwargs)
        return cls.create_action(**default)
    
    @classmethod
    def export_ooxml(cls, parent, slot, **kwargs):
        default = dict(
            text=_('To Word Processor'),
            slot=slot,
            parent=parent,
            actionicon=Icon('tango/16x16/mimetypes/x-office-document.png'),
            tip=_('Open using MS Word or Open Office')
        )
        default.update(kwargs)
        return cls.create_action(**default)
    
    @classmethod
    def backup(cls, parent, slot, **kwargs):
        default = dict(
            parent=parent,
            text=_('&Backup'),
            slot=slot,
            actionicon=cls.icon_backup,
            tip=_('Backup the database')
        )
        default.update(kwargs)
        return cls.create_action(**default)

    @classmethod
    def restore(cls, parent, slot, **kwargs):
        default = dict(
            parent=parent,
            text=_('&Restore'),
            slot=slot,
            actionicon=cls.icon_restore,
            tip=_('Restore the database from a backup')
        )
        default.update(kwargs)
        return cls.create_action(**default)

    @classmethod
    def page_setup(cls, parent, slot, **kwargs):
        default = dict(
            parent=parent,
            text=_('Page Setup...'),
            slot=slot,
            actionicon=cls.icon_pgsetup,
            tip=_('Page Setup...')
        )
        default.update(kwargs)
        return cls.create_action(**default)

    @classmethod
    def print_(cls, parent, slot, **kwargs):
        default = dict(
            parent=parent,
            text=_('Print...'),
            slot=slot,
            shortcut=QtGui.QKeySequence.Print,
            actionicon=cls.icon_print,
            tip=_('Print...')
        )
        default.update(kwargs)
        return cls.create_action(**default)

    @classmethod
    def print_preview(cls, parent, slot, **kwargs):
        default = dict(
            parent=parent,
            text=_('Print Preview'),
            slot=slot,
            actionicon=cls.icon_preview,
            tip=_('Print Preview')
        )
        default.update(kwargs)
        return cls.create_action(**default)

    @classmethod
    def exit(cls, parent, slot, **kwargs):
        default = dict(
            parent=parent,
            text=_('E&xit'),
            slot=slot,
            shortcut=QtGui.QKeySequence.Quit,
            actionicon=Icon('tango/16x16/actions/system-shutdown.png'),
            tip=_('Exit the application')
        )
        default.update(kwargs)
        return cls.create_action(**default)

    @classmethod
    def duplicate(cls, parent, slot, **kwargs):
        default = dict(
            parent=parent,
            text=_('&Copy'),
            slot=slot,
            shortcut=QtGui.QKeySequence.Copy,
            actionicon=cls.icon_copy,
            tip=_("Duplicate the selected rows")
        )
        default.update(kwargs)
        return cls.create_action(**default)

    @classmethod
    def select_all(cls, parent, slot, **kwargs):
        default = dict(
            parent=parent,
            text=_('Select &All'),
            slot=slot,
            shortcut=QtGui.QKeySequence.SelectAll,
            tip=_('Select all rows in the table'),
        )
        default.update(kwargs)
        return cls.create_action(**default)

    @classmethod
    def about(cls, parent, slot, **kwargs):
        default = dict(
            parent=parent,
            text=_('&About'),
            slot=slot,
            actionicon=Icon('tango/16x16/mimetypes/application-certificate.png'),
            tip=_("Show the application's About box")
        )
        default.update(kwargs)
        return cls.create_action(**default)

    @classmethod
    def whats_new(cls, parent, slot, **kwargs):
        default = dict(
            parent=parent,
            text=_('&What\'s new'),
            slot=slot,
            actionicon=Icon('tango/16x16/status/software-update-available.png'),
            tip=_("Show the What's New box")
        )
        default.update(kwargs)
        return cls.create_action(**default)

    @classmethod
    def affiliated_website(cls, parent, slot, **kwargs):
        default = dict(
            parent=parent,
            text=_('Affiliated website'),
            slot=slot,
            actionicon=Icon('tango/16x16/apps/internet-web-browser.png'),
            tip=_('Go to the affiliated website')
        )
        default.update(kwargs)
        return cls.create_action(**default)

    @classmethod
    def remote_support(cls, parent, slot, **kwargs):
        default = dict(
            parent=parent,
            text=_('Remote support'),
            slot=slot,
            actionicon=Icon('tango/16x16/devices/video-display.png'),
            tip=_('Let the support agent take over your desktop')
        )
        default.update(kwargs)
        return cls.create_action(**default)

    @classmethod
    def help(cls, parent, slot, **kwargs):
        default = dict(
            parent=parent,
            text=_('Help'),
            slot=slot,
            shortcut=QtGui.QKeySequence.HelpContents,
            actionicon=cls.icon_help,
            tip=_('Help content')
        )
        default.update(kwargs)
        return cls.create_action(**default)

    @classmethod
    def new(cls, parent, slot, **kwargs):
        default = dict(
            parent=parent,
            text=_('New'),
            slot=slot,
            shortcut=QtGui.QKeySequence.New,
            actionicon=cls.icon_new,
            tip=_('New')
        )
        default.update(kwargs)
        return cls.create_action(**default)

    @classmethod
    def delete(cls, parent, slot, **kwargs):
        default = dict(
            parent=parent,
            text=_('Delete'),
            slot=slot,
            shortcut=QtGui.QKeySequence.Delete,
            actionicon=cls.icon_delete,
            tip=_('Delete')
        )
        default.update(kwargs)
        return cls.create_action(**default)
 
    @classmethod
    def update_values(cls, parent, slot, **kwargs):
        default = dict(
            parent = parent,
            text = _('Replace field contents'),
            slot = slot,
            tip = _('Replace the content of a field for all rows in a selection')
        )
        default.update(kwargs)
        return cls.create_action(**default)

    @classmethod
    def merge_document(cls, parent, slot, **kwargs):
        default = dict(
            parent = parent,
            text = _('Merge document'),
            slot = slot,
            tip = _('Merge a template document with all rows in a selection')
        )
        default.update(kwargs)
        return cls.create_action(**default)

    @classmethod
    def export_excel(cls, parent, slot, **kwargs):
        default = dict(
            parent=parent,
            text=_('Export to MS Excel'),
            slot=slot,
            actionicon=cls.icon_excel,
            tip=_('Export to MS Excel')
        )
        default.update(kwargs)
        return cls.create_action(**default)

    @classmethod
    def export_word(cls, parent, slot, **kwargs):
        default = dict(
            parent=parent,
            text=_('Export to MS Word'),
            slot=slot,
            actionicon=cls.icon_word,
            tip=_('Export to MS Word')
        )
        default.update(kwargs)
        return cls.create_action(**default)

    @classmethod
    def export_mail(cls, parent, slot, **kwargs):
        default = dict(
            parent=parent,
            text=_('Send by e-mail'),
            slot=slot,
            actionicon=cls.icon_mail,
            tip=_('Send by e-mail')
        )
        default.update(kwargs)
        return cls.create_action(**default)

    @classmethod
    def import_file(cls, parent, slot, **kwargs):
        default = dict(
            parent=parent,
            text=_('Import from file'),
            slot=slot,
            actionicon=cls.icon_import,
            tip=_('Import from file')
        )
        default.update(kwargs)
        return cls.create_action(**default)
        
    @classmethod
    def refresh(cls, parent, slot, **kwargs):
        default = dict(
            parent = parent,
            text = _('Refresh'),
            slot = slot,
            shortcut = QtGui.QKeySequence( Qt.Key_F9 ),
            icond = Icon('tango/16x16/actions/view-refresh.png'),
        )
        default.update(kwargs)
        return cls.create_action(**default)

    @classmethod
    def new_tab(cls, parent, slot, **kwargs):
        default = dict(
            parent = parent,
            text = _('Open in New Tab'),
            slot = slot
        )
        default.update(kwargs)
        return cls.create_action(**default)



#  ============================================================================
#
#  Copyright (C) 2007-2011 Conceptive Engineering bvba. All rights reserved.
#  www.conceptive.be / project-camelot@conceptive.be
#
#  This file is part of the Camelot Library.
#
#  This file may be used under the terms of the GNU General Public
#  License version 2.0 as published by the Free Software Foundation
#  and appearing in the file license.txt included in the packaging of
#  this file.  Please review this information to ensure GNU
#  General Public Licensing requirements will be met.
#
#  If you are unsure which license is appropriate for your use, please
#  visit www.python-camelot.com or contact project-camelot@conceptive.be
#
#  This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
#  WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
#
#  For use of this library in commercial applications, please contact
#  project-camelot@conceptive.be
#
#  ============================================================================
from camelot.container import Container

class FigureContainer( Container ):
    """A container that is able to plot itself on a matplotlib figure canvas.
    
    Its 'plot_on_figure' method will be called in the gui thread to fill the figure
    canvas.
    
    One figure canvas can contain multiple axes (=sub plots)
    """
    
    def __init__(self, axes):
        """
        :param axes: a list of AxesContainer objects representing all the subplots, in
        the form of ::
        
          [[ax1, ax2],
           [ax3, ax4]]
           
        """
        self.axes = axes
        
    def plot_on_figure(self, fig):
        """Draw all axes (sub plots) on a figure canvas"""
        fig.clear()
        if self.axes:
            rows = len(self.axes)
            cols = len(self.axes[0])
            for i,row in enumerate(self.axes):
                for j,subplot in enumerate(row):
                    n = i*cols + j
                    ax = fig.add_subplot( rows, cols, n+1 )
                    ax.clear()
                    subplot.plot_on_axes( ax )
                
class AxesMethod(object):
    """Helper class to substitute a method on an Axes object and
    record its calls"""
    
    def __init__(self, method_name, commands):
        """
        :param method_name: the name of the method for which this object is a substitute
        :param commands: a list in which to store invocations of the method
        """
        self._method_name = method_name
        self._commands = commands
        
    def __call__(self, *args, **kwargs):
        """record a call the the substitute method into the commands list"""
        self._commands.append( (self._method_name, args, kwargs) )
            
class AxesContainer( Container ):
    """A container that is able to generate a plot on a matplotlib axes.  Methods
    can be called on this class as if it were a matplotlib Axes class.  All method
    calls will be recorded.  Of course the methods won't return matplotlib objects.
    The set_auto_legend method can be used to turn legens on without the need for
    matplotlib objects.
    """

    def __init__(self):
        """
        :param legend: True or False, to put a legend on the chart
        """
        super(AxesContainer, self).__init__()
        # store all the method calls that need to be called on a
        # matplotlib axes object in a list
        self._commands = list()
        self._auto_legend = False
        
    def __getattr__(self, attribute_name):
        """Suppose the caller wants to call a function on a matplotlib
        axes object"""
        return AxesMethod( attribute_name, self._commands )
       
    def set_auto_legend(self, auto_legend):
        """Specify if the container should try to put a legend on the 
        plot.
        :param auto_legend: True or False
        """
        self._auto_legend = auto_legend
  
    def plot_on_axes(self, ax):
        """Replay the list of stored commands to the real Axes object"""
        for name, args, kwargs in self._commands:
            getattr(ax, name)(*args, **kwargs)
        if self._auto_legend:
            handles, labels = ax.get_legend_handles_labels()
            ax.legend(handles, labels)


class PlotContainer( AxesContainer ):

    # this line drives pylint crazy because it need axes to be imported
    #__doc__ = axes.Axes.plot.__doc__
    
    def __init__(self, *args, **kwargs):
        """:param *args, **kwargs: the arguments to be passed to the matplotlib plot command"""
        super(PlotContainer, self).__init__()
        self.plot( *args, **kwargs )
        
class BarContainer( AxesContainer ):

    # this line drives pylint crazy because it need axes to be imported
    #__doc__ = axes.Axes.bar.__doc__
    
    def __init__(self, *args, **kwargs):
        """:param *args, **kwargs: the arguments to be passed to the matplotlib bar command"""
        super(BarContainer, self).__init__()
        self.bar( *args, **kwargs )
                    
def structure_to_figure_container( structure ):
    """Convert a structure to a figure container, if the structure
    is an instance of a FigureContainer, return as is.
    
    If the structure is an instance of an AxesContainer, return a
    FigureContainer with a single Axes.
    
    If the structure is a list, use the structure as a constructor
    argument for the FigureContainer
    """
    
    if isinstance(structure, FigureContainer):
        return structure
    if isinstance(structure, AxesContainer):
        return FigureContainer( [[structure]] )
    if isinstance(structure, (list, tuple)):
        return FigureContainer( structure )





#  ============================================================================
#
#  Copyright (C) 2007-2011 Conceptive Engineering bvba. All rights reserved.
#  www.conceptive.be / project-camelot@conceptive.be
#
#  This file is part of the Camelot Library.
#
#  This file may be used under the terms of the GNU General Public
#  License version 2.0 as published by the Free Software Foundation
#  and appearing in the file license.txt included in the packaging of
#  this file.  Please review this information to ensure GNU
#  General Public Licensing requirements will be met.
#
#  If you are unsure which license is appropriate for your use, please
#  visit www.python-camelot.com or contact project-camelot@conceptive.be
#
#  This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
#  WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
#
#  For use of this library in commercial applications, please contact
#  project-camelot@conceptive.be
#
#  ============================================================================
from camelot.model import metadata
from elixir.entity import Entity
from elixir.options import using_options
from elixir.fields import Field
from sqlalchemy.types import Unicode, INT
"""Classes to support the loading of required datasets into the 
database"""


__metadata__ = metadata

class Fixture( Entity ):
    """Keep track of static data loaded into the database"""
    using_options( tablename = 'fixture' )
    model = Field( Unicode( 256 ), index = True, required = True )
    primary_key = Field( INT(), index = True, required = True )
    fixture_key = Field( Unicode( 256 ), index = True, required = True )
    fixture_class = Field( Unicode( 256 ), index = True, required = False )

    @classmethod
    def findFixtureReference( cls, entity, fixture_key, fixture_class = None ):
        entity_name = unicode( entity.__name__ )
        return cls.query.filter_by( model = unicode( entity_name ), fixture_key = fixture_key, fixture_class = fixture_class ).first()

    @classmethod
    def findFixture( cls, entity, fixture_key, fixture_class = None ):
        """Find a registered fixture, return None if no fixture is found"""
        reference = cls.findFixtureReference( entity, fixture_key, fixture_class )
        if reference:
            return entity.get( reference.primary_key )

    @classmethod
    def findFixtureKey( cls, entity, primary_key ):
        """Find the fixture key for an object of type entity with primary key
        :return: fixture_key        
        """
        entity_name = unicode( entity.__name__ )
        fixture = cls.query.filter_by( model = entity_name, primary_key = primary_key ).first()
        if fixture:
            return fixture.fixture_key
        else:
            return None
        
    @classmethod
    def findFixtureKeyAndClass( cls, obj ):
        """Find the fixture key and class of an object
        @param obj: the object we are looking for 
        @return: (fixture_key, fixture_class) if the object is a registered fixture, (None, None) otherwise
        """
        entity_name = unicode( obj.__class__.__name__ )
        fixture = cls.query.filter_by( model = entity_name, primary_key = obj.id ).first()
        if fixture:
            return ( fixture.fixture_key, fixture.fixture_class )
        else:
            return ( None, None )
        
    @classmethod
    def findFixtureKeysAndClasses( cls, entity ):
        """Load all fixture keys of a certain entity in batch
        :param entity: the model class for which the fixtures should be found
        :return: a dictionary mapping the primary key of a on object of type entity to its (fixture key, fixture class)
        """
        entity_name = unicode( entity.__name__ )
        return dict((fixture.primary_key, (fixture.fixture_key, fixture.fixture_class)) for fixture in cls.query.filter_by( model = entity_name ).all())

    @classmethod
    def insertOrUpdateFixture( cls, entity, fixture_key, values, fixture_class = None ):
        from sqlalchemy.orm.session import Session
        obj = cls.findFixture( entity, fixture_key, fixture_class )
        store_fixture = False
        if not obj:
            obj = entity()
            store_fixture = True
        obj.from_dict( values )
        Session.object_session( obj ).flush( [obj] )
        if store_fixture:
            #
            # The fixture itself might have been deleted, but the reference might be intact,
            # so this should be updated
            #
            reference = cls.findFixtureReference( entity, fixture_key, fixture_class )
            if not reference:
                reference = cls( model = unicode( entity.__name__ ), primary_key = obj.id, fixture_key = fixture_key, fixture_class = fixture_class )
            else:
                reference.primary_key = obj.id
            Session.object_session( reference ).flush( [reference] )
        return obj
    
    @classmethod
    def removeAllFixtures( cls, entity ):
        for fixture_key, fixture_class in cls.findFixtureKeysAndClasses( entity ).values():
            cls.removeFixture(entity, fixture_key, fixture_class)
            
    @classmethod
    def removeFixture( cls, entity, fixture_key, fixture_class ):
        """Remove a fixture from the database"""
        # remove the object itself
        from sqlalchemy.orm.session import Session
        obj = cls.findFixture( entity, fixture_key, fixture_class)
        print 'remove', unicode(obj)
        obj.delete()
        Session.object_session( obj ).flush( [obj] )
        # if this succeeeds, remove the reference
        reference = cls.findFixtureReference(entity, fixture_key, fixture_class)
        reference.delete()
        Session.object_session( reference ).flush( [reference] )
        

class FixtureVersion( Entity ):
    """Keep track of the version the fixtures have in the current database, the subversion
    revision number is a good candidate to be used as a fixture version.
    
    :return: an integer representing the current version, 0 if no version found
    """
    using_options( tablename = 'fixture_version' )
    fixture_version = Field( INT(), index = True, required = True, default=0 )
    fixture_class = Field( Unicode( 256 ), index = True, required = False, unique=True )    
    
    @classmethod
    def get_current_version( cls, fixture_class = None ):
        """Get the current version of the fixtures in the database for a certain 
        fixture class.
        
        :param fixture_class: the fixture class for which to get the version
        """
        obj = cls.query.filter_by( fixture_class = fixture_class ).first()
        if obj:
            return obj.fixture_version
        return 0
        
    @classmethod
    def set_current_version( cls, fixture_class = None, fixture_version = 0 ):
        """Set the current version of the fixtures in the database for a certain 
        fixture class.
        
        :param fixture_class: the fixture class for which to get the version
        :param fixture_version: the version number to which to set the fixture version
        """
        from sqlalchemy.orm.session import Session
        obj = cls.query.filter_by( fixture_class = fixture_class ).first()
        if not obj:
            obj = FixtureVersion( fixture_class = fixture_class )
        obj.fixture_version = fixture_version
        Session.object_session( obj ).flush( [obj] ) 




#  ============================================================================
#
#  Copyright (C) 2007-2011 Conceptive Engineering bvba. All rights reserved.
#  www.conceptive.be / project-camelot@conceptive.be
#
#  This file is part of the Camelot Library.
#
#  This file may be used under the terms of the GNU General Public
#  License version 2.0 as published by the Free Software Foundation
#  and appearing in the file license.txt included in the packaging of
#  this file.  Please review this information to ensure GNU
#  General Public Licensing requirements will be met.
#
#  If you are unsure which license is appropriate for your use, please
#  visit www.python-camelot.com or contact project-camelot@conceptive.be
#
#  This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
#  WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
#
#  For use of this library in commercial applications, please contact
#  project-camelot@conceptive.be
#
#  ============================================================================

import logging
logger = logging.getLogger('camelot.view.controls.delegates.delegatemanager')

from PyQt4 import QtGui, QtCore
from PyQt4.QtCore import Qt

class DelegateManager(QtGui.QItemDelegate):
    """Manages custom delegates, should not be used by the application
  developer
  """

    def __init__(self, parent=None, **kwargs):
        QtGui.QItemDelegate.__init__(self, parent)
        self.delegates = {}

    def set_columns_desc(self, columnsdesc):
        self.columnsdesc = columnsdesc

    def get_column_delegate(self, column):
        try:
            return self.delegates[column]
        except KeyError:
            logger.error('Programming Error, no delegate available for column %s'%column)
            logger.error('Available columns : %s'%unicode(self.delegates.keys()))
            raise KeyError

    def insertColumnDelegate(self, column, delegate):
        """Inserts a custom column delegate"""
        logger.debug('inserting delegate for column %s' % column)
        assert delegate != None
        delegate.setParent(self)
        self.delegates[column] = delegate
        delegate.commitData.connect(self._commit_data)
        delegate.closeEditor.connect(self._close_editor)

    def _commit_data(self, editor):
        self.commitData.emit(editor)

    @QtCore.pyqtSlot( QtGui.QWidget, QtGui.QAbstractItemDelegate.EndEditHint )
    def _close_editor(self, editor, hint):
        self.closeEditor.emit(editor, hint )

    def removeColumnDelegate(self, column):
        """Removes custom column delegate"""
        logger.debug('removing a custom column delegate')
        if column in self.delegates:
            del self.delegates[column]

    def paint(self, painter, option, index):
        """Use a custom delegate paint method if it exists"""
        delegate = self.get_column_delegate(index.column())
        delegate.paint(painter, option, index)

    def createEditor(self, parent, option, index):
        """Use a custom delegate createEditor method if it exists"""
        delegate = self.get_column_delegate(index.column())
        editor = delegate.createEditor(parent, option, index)
        assert editor != None
        return editor

    def setEditorData(self, editor, index):
        """Use a custom delegate setEditorData method if it exists"""
        logger.debug('setting editor data for column %s' % index.column())
        try:
            delegate = self.get_column_delegate(index.column())
            delegate.setEditorData(editor, index)
        except Exception, e:
            logger.error('Programming Error : could not set editor data for editor at column %s'%(index.column()), exc_info=e)
            logger.error('value that could not be set : %s'%unicode(index.model().data(index, Qt.EditRole)))

    def setModelData(self, editor, model, index):
        """Use a custom delegate setModelData method if it exists"""
        logger.debug('setting model data for column %s' % index.column())
        delegate = self.get_column_delegate(index.column())
        delegate.setModelData(editor, model, index)

    def sizeHint(self, option, index):
        option = QtGui.QStyleOptionViewItem()
        delegate = self.get_column_delegate(index.column())
        return delegate.sizeHint(option, index)





import logging

logging.basicConfig(level=logging.INFO, format='[%(levelname)-7s] [%(name)-35s] - %(message)s')
#logging.getLogger('sqlalchemy.engine').setLevel(logging.INFO)

def setup_model():
    import camelot.model
    import camelot_example.model
    from camelot_example.view import setup_views
    from camelot_example.fixtures import load_movie_fixtures
    from elixir import setup_all
    from camelot.model.authentication import updateLastLogin
    setup_all(create_tables=True)
    setup_views()
    load_movie_fixtures()
    updateLastLogin()

CAMELOT_MEDIA_ROOT = 'media'

def ENGINE():
   from sqlalchemy import create_engine
   return create_engine( 'sqlite:///' )




from camelot.test import ModelThreadTestCase

class ModelCase( ModelThreadTestCase ):
  """Test the build in camelot model"""
  
  def setUp(self):
    super( ModelCase, self ).setUp()
    from camelot.model.authentication import Person
    from camelot.view.proxy.queryproxy import QueryTableProxy
    from camelot.admin.application_admin import ApplicationAdmin
    self.app_admin = ApplicationAdmin()
    self.person_admin = self.app_admin.get_related_admin( Person )
  
  def test_person_contact_mechanism( self ):
    from camelot.model.authentication import Person
    person = Person( first_name = u'Robin',
                     last_name = u'The brave' )
    self.assertEqual( person.contact_mechanisms_email, None )
    mechanism = ('email', 'robin@test.org')
    person.contact_mechanisms_email = mechanism
    self.person_admin.flush( person )
    self.assertEqual( person.contact_mechanisms_email, mechanism )
    self.person_admin.delete( person )
    person = Person( first_name = u'Robin',
                     last_name = u'The brave' )
    self.person_admin.flush( person )
    self.assertEqual( person.contact_mechanisms_email[1], u'' )
    


#  ============================================================================
#
#  Copyright (C) 2007-2011 Conceptive Engineering bvba. All rights reserved.
#  www.conceptive.be / project-camelot@conceptive.be
#
#  This file is part of the Camelot Library.
#
#  This file may be used under the terms of the GNU General Public
#  License version 2.0 as published by the Free Software Foundation
#  and appearing in the file license.txt included in the packaging of
#  this file.  Please review this information to ensure GNU
#  General Public Licensing requirements will be met.
#
#  If you are unsure which license is appropriate for your use, please
#  visit www.python-camelot.com or contact project-camelot@conceptive.be
#
#  This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
#  WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
#
#  For use of this library in commercial applications, please contact
#  project-camelot@conceptive.be
#
#  ============================================================================

import logging

from PyQt4 import QtCore

from camelot.admin.action import ActionStep
from camelot.core.exception import GuiException, CancelRequest
from camelot.view.model_thread import post

LOGGER = logging.getLogger( 'camelot.view.action_runner' )

class ActionRunner( QtCore.QEventLoop ):
    """Helper class for handling the signals and slots when an action
    is running.  This class takes a generator and iterates it within the
    model thread while taking care of Exceptions raised and ActionSteps
    yielded by the generator.
    
    This is class is intended for internal Camelot use only.
    """
    
    non_blocking_action_step_signal = QtCore.pyqtSignal(object)
    
    def __init__( self, generator_function, gui_context ):
        """
        :param generator_function: function to be called in the model thread,
            that will return the generator
        :param gui_context: the GUI context of the generator
        """
        super( ActionRunner, self ).__init__()
        self._return_code = None
        self._generator_function = generator_function
        self._generator = None
        self._gui_context = gui_context
        self._model_context = gui_context.create_model_context()
        self.non_blocking_action_step_signal.connect( self.non_blocking_action_step )
        post( self._initiate_generator, self.generator, self.exception )
    
    def exit( self, return_code = 0 ):
        """Reimplementation of exit to store the return code"""
        self._return_code = return_code
        return super( ActionRunner, self ).exit( return_code )
    
    def exec_( self, flags = QtCore.QEventLoop.AllEvents ):
        """Reimplementation of exec_ to prevent the event loop being started
        when exit has been called prior to calling exec_.
        
        This can be the case when running in single threaded mode.
        """
        if self._return_code == None:
            return super( ActionRunner, self ).exec_( flags )
        return self._return_code
        
    def _initiate_generator( self ):
        """Create the model context and start the generator"""
        return self._generator_function( self._model_context )
            
    def _iterate_until_blocking( self, generator_method, *args ):
        """Helper calling for generator methods.  The decorated method iterates
        the generator until the generator yields an :class:`ActionStep` object that
        is blocking.  If a non blocking :class:`ActionStep` object is yielded, then
        send it to the GUI thread for execution through the signal slot mechanism.
        
        :param generator_method: the method of the generator to be called
        :param *args: the arguments to use when calling the generator method.
        """
        try:
            result = generator_method( *args )
            while True:
                if isinstance( result, (ActionStep,)):
                    if result.blocking:
                        return result
                    else:
                        self.non_blocking_action_step_signal.emit( result )
                result = self._generator.next()
        except CancelRequest, e:
            return e
        except StopIteration, e:
            return e

    @QtCore.pyqtSlot( object )
    def non_blocking_action_step( self, action_step ):
        action_step.gui_run( self._gui_context )
        
    @QtCore.pyqtSlot( object )
    def exception( self, exception_info ):
        """Handle an exception raised by the generator"""
        from camelot.view.controls.exception import model_thread_exception_message_box
        model_thread_exception_message_box( exception_info )
        self.exit()
        
    @QtCore.pyqtSlot( object )
    def generator( self, generator ):
        """Handle the creation of the generator"""
        self._generator = generator
        #
        # when model_run is not a generator, but a normal function it returns
        # no generator, and as such we can exit the event loop
        #
        if self._generator != None:
            post( self._iterate_until_blocking, 
                  self.next, 
                  self.exception,
                  args = ( self._generator.next, ) )
        else:
            self.exit()
        
    def _was_canceled( self, gui_context ):
        """raise a :class:`camelot.core.exception.CancelRequest` if the
        user pressed the cancel button of the progress dialog in the
        gui_context.
        """
        if gui_context.progress_dialog:
            if gui_context.progress_dialog.wasCanceled():
                raise CancelRequest()
            
    @QtCore.pyqtSlot( object )
    def next( self, yielded ):
        """Handle the result of the next call of the generator
        
        :param yielded: the object that was yielded by the generator in the
            *model thread*
        """
        if isinstance( yielded, (ActionStep,) ):
            try:
                self._was_canceled( self._gui_context )
                to_send = yielded.gui_run( self._gui_context )
                self._was_canceled( self._gui_context )
                post( self._iterate_until_blocking, 
                      self.next, 
                      self.exception, 
                      args = ( self._generator.send, to_send,) )
            except CancelRequest, exc:
                post( self._iterate_until_blocking,
                      self.next,
                      self.exception,
                      args = ( self._generator.throw, exc,) )
            except Exception, exc:
                LOGGER.error( 'gui exception while executing action', 
                              exc_info=exc)
                #
                # In case of an exception in the GUI thread, propagate an
                # exception to make sure the generator ends.  Don't propagate
                # the very same exception, because no references from the GUI
                # should be past to the model.
                #
                post( self._iterate_until_blocking,
                      self.next,
                      self.exception,
                      args = ( self._generator.throw, GuiException(), ) )
        elif isinstance( yielded, (StopIteration, CancelRequest) ):
            #
            # Process the events before exiting, as there might be exceptions
            # left in the signal slot queue
            #
            self.processEvents()
            self.exit()
        else:
            LOGGER.error( 'next call of generator returned an unexpected object of type %s'%( yielded.__class__.__name__ ) ) 
            LOGGER.error( unicode( yielded ) )
            raise Exception( 'this should not happen' )



#  ============================================================================
#
#  Copyright (C) 2007-2011 Conceptive Engineering bvba. All rights reserved.
#  www.conceptive.be / project-camelot@conceptive.be
#
#  This file is part of the Camelot Library.
#
#  This file may be used under the terms of the GNU General Public
#  License version 2.0 as published by the Free Software Foundation
#  and appearing in the file license.txt included in the packaging of
#  this file.  Please review this information to ensure GNU
#  General Public Licensing requirements will be met.
#
#  If you are unsure which license is appropriate for your use, please
#  visit www.python-camelot.com or contact project-camelot@conceptive.be
#
#  This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
#  WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
#
#  For use of this library in commercial applications, please contact
#  project-camelot@conceptive.be
#
#  ============================================================================

"""Controls to filter data"""

import logging
logger = logging.getLogger('camelot.view.controls.filter')

from PyQt4 import QtGui, QtCore

class FilterList( QtGui.QWidget ):
    """A list with filters that can be applied on a query in the tableview"""

    filters_changed_signal = QtCore.pyqtSignal()
    
    def __init__(self, items, parent):
        """
    :param items: list of tuples (filter, (name, choices)) for constructing the different filterboxes
    """
        super(FilterList, self).__init__(parent)
        #self.setFrameStyle(QtGui.QFrame.NoFrame)
        layout = QtGui.QVBoxLayout()
        layout.setSpacing( 4 )
        for filter, (name, options) in items:
            filter_widget = filter.render(self, name, options)
            layout.addWidget(filter_widget)
            filter_widget.filter_changed_signal.connect( self.emit_filters_changed )
        layout.addStretch()
        self.setLayout(layout)
        if len(items) == 0:
            self.setMaximumWidth(0)
        else:
            self.setSizePolicy( QtGui.QSizePolicy.MinimumExpanding, 
                                QtGui.QSizePolicy.Expanding )

    def decorate_query(self, query):
        for i in range(self.layout().count()):
            if self.layout().itemAt(i).widget():
                query = self.layout().itemAt(i).widget().decorate_query(query)
        return query

    @QtCore.pyqtSlot()
    def emit_filters_changed(self):
        logger.debug('filters changed')
        self.filters_changed_signal.emit()



#  ============================================================================
#
#  Copyright (C) 2007-2011 Conceptive Engineering bvba. All rights reserved.
#  www.conceptive.be / project-camelot@conceptive.be
#
#  This file is part of the Camelot Library.
#
#  This file may be used under the terms of the GNU General Public
#  License version 2.0 as published by the Free Software Foundation
#  and appearing in the file license.txt included in the packaging of
#  this file.  Please review this information to ensure GNU
#  General Public Licensing requirements will be met.
#
#  If you are unsure which license is appropriate for your use, please
#  visit www.python-camelot.com or contact project-camelot@conceptive.be
#
#  This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
#  WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
#
#  For use of this library in commercial applications, please contact
#  project-camelot@conceptive.be
#
#  ============================================================================

import logging

from PyQt4 import QtGui
from PyQt4 import QtCore

from camelot.view.controls.editors.customeditor import CustomEditor
from camelot.view.controls.editors.wideeditor import WideEditor
from camelot.view.proxy import ValueLoading
from camelot.view.art import Icon
from camelot.core.utils import ugettext as _

PAD_INCHES = 0.1

LOGGER = logging.getLogger('camelot.view.controls.editors.charteditor')

class ChartEditor(QtGui.QFrame, CustomEditor, WideEditor):
    """Editor to display and manipulate matplotlib charts.  The editor
    itself is generic for all kinds of plots,  it simply provides the
    data to be ploted with a set of axes.  The data itself should know
    how exactly to plot itself.
    """

    show_fullscreen_signal = QtCore.pyqtSignal()
    editingFinished = QtCore.pyqtSignal()

    def __init__(self, parent=None, width=50, height=40, dpi=50, field_name='chart', **kwargs):
        from matplotlib.figure import Figure
        from matplotlib.backends.backend_qt4agg import FigureCanvasQTAgg as FigureCanvas
        super(ChartEditor, self).__init__(parent)
        self.setObjectName( field_name )
        
        chart_frame = QtGui.QFrame( self )
        chart_frame.setFrameShape( self.Box )
        chart_frame.setContentsMargins( 1, 1, 1, 1 )
        chart_frame_layout = QtGui.QHBoxLayout()
        chart_frame_layout.setContentsMargins( 0, 0, 0, 0)
        chart_frame.setLayout( chart_frame_layout )

        # find out background color, because using a transparent
        # figure fails when the window is resized: the background
        # is not redrawn
        # need to do str() else matplotlib gets confused by the qstring
        # bgcolorrgb = str(self.palette().background().color().name())
        self.fig = Figure(
            figsize=(width, height),
            dpi=dpi,
            facecolor='#ffffff',
        )
        layout = QtGui.QHBoxLayout()
        self.canvas = FigureCanvas( self.fig )
        chart_frame_layout.addWidget( self.canvas )
        layout.addWidget(chart_frame)
        button_layout = QtGui.QVBoxLayout()
        button_layout.setSpacing( 0 )

        icon = Icon( 'tango/16x16/actions/document-print-preview.png' ).getQIcon()
        button_layout.addStretch()
        
        print_button = QtGui.QToolButton()
        print_button.setIcon( icon )
        print_button.setAutoRaise( True )
        print_button.setToolTip( _('Print Preview') )
        print_button.clicked.connect( self.print_preview )
        button_layout.addWidget( print_button )

        icon = Icon( 'tango/16x16/actions/edit-copy.png' ).getQIcon()
        copy_button = QtGui.QToolButton()
        copy_button.setIcon( icon )
        copy_button.setAutoRaise( True )
        copy_button.setToolTip( _('Copy to clipboard') )
        copy_button.clicked.connect( self.copy_to_clipboard )
        button_layout.addWidget( copy_button )
                
        layout.addLayout( button_layout )
        layout.setContentsMargins( 0, 0, 0, 0)
        self.setLayout(layout)
        self.canvas.setSizePolicy(
            QtGui.QSizePolicy.Expanding,
            QtGui.QSizePolicy.Expanding
        )
        self.canvas.installEventFilter(self)
        self.show_fullscreen_signal.connect(self.show_fullscreen)
        self.canvas.updateGeometry()
        self._litebox = None

    @QtCore.pyqtSlot()
    def copy_to_clipboard(self):
        """Copy the chart to the clipboard"""
        clipboard = QtGui.QApplication.clipboard()
        pixmap = QtGui.QPixmap.grabWidget( self.canvas )
        clipboard.setPixmap( pixmap )
        
    @QtCore.pyqtSlot()
    def print_preview(self):
        """Popup a print preview dialog for the Chart"""
        dialog = QtGui.QPrintPreviewDialog()            
        dialog.paintRequested.connect( self.on_paint_request )
        dialog.exec_()
        
    @QtCore.pyqtSlot( QtGui.QPrinter )
    def on_paint_request(self, printer):
        from matplotlib.figure import Figure
        from matplotlib.backends.backend_qt4agg import FigureCanvasQTAgg as FigureCanvas
        rect = printer.pageRect( QtGui.QPrinter.Inch )
        dpi = printer.resolution()
        fig = Figure( facecolor='#ffffff')
        fig.set_figsize_inches( (rect.width(),rect.height()) )
        fig.set_dpi( dpi )
        self._value.plot_on_figure( fig )
        canvas = FigureCanvas(fig)
        canvas.render( printer )
    
    def set_field_attributes(self, *args, **kwargs):
        """Overwrite set_field attributes because a ChartEditor cannot be disabled
        or have its background color changed"""
        pass
    
    @staticmethod
    def show_fullscreen_chart(chart, parent):
        """
        :param chart: a chart container
        :return: the widget showing the chart, by default a LiteBoxView
        """
        from matplotlib.figure import Figure
        from matplotlib.backends.backend_qt4agg import FigureCanvasQTAgg as FigureCanvas
        from camelot.view.controls.liteboxview import LiteBoxView
        from camelot.container.chartcontainer import structure_to_figure_container
        figure_container = structure_to_figure_container( chart )
        litebox = LiteBoxView(parent)
        fig = Figure(facecolor='#ffffff')
        canvas = FigureCanvas(fig)
        canvas.updateGeometry()
        figure_container.plot_on_figure(fig)
        proxy = QtGui.QGraphicsProxyWidget()
        proxy.setWidget(canvas)
        litebox.show_fullscreen_item(proxy)
        canvas.draw()
        return litebox
                    
    @QtCore.pyqtSlot()
    def show_fullscreen(self):
        """Show the plot full screen, using the litebox"""
        if self._value:
            # if we give the litebox a parent, the close button does not work ??
            self._litebox = self.show_fullscreen_chart(self._value, None)

    def eventFilter(self, object, event):
        """intercept mouse clicks on a chart to show the chart fullscreen"""
        if not object.isWidgetType():
            return False
        if event.type() != QtCore.QEvent.MouseButtonPress:
            return False
        if event.modifiers() != QtCore.Qt.NoModifier:
            return False
        if event.buttons() == QtCore.Qt.LeftButton:
            self.show_fullscreen_signal.emit()
            return True
        return False

    def set_value(self, value):
        """Accepts a camelot.container.chartcontainer.FigureContainer or a 
        camelot.container.chartcontainer.AxesContainer """
        from camelot.container.chartcontainer import structure_to_figure_container
        self._value = structure_to_figure_container( super(ChartEditor, self).set_value(value) )
        self.on_draw()

#    def _get_tightbbox(self, fig, pad_inches):
#        renderer = fig.canvas.get_renderer()
#        bbox_inches = fig.get_tightbbox(renderer)
#        return bbox_inches.padded(pad_inches)
#
#    def tight_borders(self, fig, pad_inches=PAD_INCHES):
#        """Stretch subplot boundaries to figure edges plus padding."""
#        # call draw to update the renderer and get accurate bboxes.
#        import numpy as np
#        fig.canvas.draw()
#        bbox_original = fig.bbox_inches
#        bbox_tight = self._get_tightbbox(fig, pad_inches)
#        
#        print bbox_tight
#        
#        # figure dimensions ordered like bbox.extents: x0, y0, x1, y1
#        lengths = np.array([bbox_original.width, bbox_original.height,
#                            bbox_original.width, bbox_original.height])
#        whitespace = (bbox_tight.extents - bbox_original.extents) / lengths
#        
#        # border padding ordered like bbox.extents: x0, y0, x1, y1
#        current_borders = np.array([fig.subplotpars.left, fig.subplotpars.bottom,
#                                    fig.subplotpars.right, fig.subplotpars.top])
#        
#        
#        left, bottom, right, top = current_borders - whitespace
#        print self.canvas.size()
#        print left, bottom, right, top
#        fig.subplots_adjust(bottom=bottom, top=top, left=left, right=right)
    
    def on_draw(self):
        """draw the matplotlib figure on the canvas"""
        if self._value not in (None, ValueLoading):
            self._value.plot_on_figure(self.fig)
            self.canvas.draw()
#        renderer = self.fig.canvas.get_renderer()
#        print 'widget size', self.canvas.size()
#        print 'renderer size', renderer.get_canvas_width_height()
#        print 'points_to_pixels', renderer.points_to_pixels(1.0)
#        print 'tightbbox', self.fig.get_tightbbox(renderer)
#        #self.fig.subplots_adjust(bottom=0.3, right=0.9, top=0.9, left=0.1)
#        self.tight_borders(self.fig)
#        self.canvas.draw()






#  ============================================================================
#
#  Copyright (C) 2007-2011 Conceptive Engineering bvba. All rights reserved.
#  www.conceptive.be / project-camelot@conceptive.be
#
#  This file is part of the Camelot Library.
#
#  This file may be used under the terms of the GNU General Public
#  License version 2.0 as published by the Free Software Foundation
#  and appearing in the file license.txt included in the packaging of
#  this file.  Please review this information to ensure GNU
#  General Public Licensing requirements will be met.
#
#  If you are unsure which license is appropriate for your use, please
#  visit www.python-camelot.com or contact project-camelot@conceptive.be
#
#  This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
#  WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
#
#  For use of this library in commercial applications, please contact
#  project-camelot@conceptive.be
#
#  ============================================================================

from PyQt4 import QtGui
from customeditor import AbstractCustomEditor, draw_tooltip_visualization


class TextLineEditor(QtGui.QLineEdit, AbstractCustomEditor):

    def __init__(self, 
                 parent, 
                 length = 20, 
                 field_name = 'text_line',
                 **kwargs):
        QtGui.QLineEdit.__init__(self, parent)
        self.setObjectName( field_name )
        AbstractCustomEditor.__init__(self)
        if length:
            self.setMaxLength(length)

    def set_value(self, value):
        value = AbstractCustomEditor.set_value(self, value)
        if value is not None:
            self.setText(unicode(value))
        else:
            self.setText('')
        return value

    def get_value(self):
        value_loading = AbstractCustomEditor.get_value(self)
        if value_loading is not None:
            return value_loading

        value = unicode(self.text())
        if self.value_is_none and not value:
            return None

        return value

    def set_field_attributes(self, editable = True,
                                   background_color = None, 
                                   tooltip = None, **kwargs):
        self.set_background_color( background_color )
        self.set_enabled( editable )
        self.setToolTip(unicode(tooltip or ''))

    def set_enabled(self, editable=True):
        value = self.text()
        self.setEnabled(editable)
        self.setText(value)

    def paintEvent(self, event):
        super(TextLineEditor, self).paintEvent(event)
        
        if self.toolTip():
            draw_tooltip_visualization(self)



#  ============================================================================
#
#  Copyright (C) 2007-2011 Conceptive Engineering bvba. All rights reserved.
#  www.conceptive.be / project-camelot@conceptive.be
#
#  This file is part of the Camelot Library.
#
#  This file may be used under the terms of the GNU General Public
#  License version 2.0 as published by the Free Software Foundation
#  and appearing in the file license.txt included in the packaging of
#  this file.  Please review this information to ensure GNU
#  General Public Licensing requirements will be met.
#
#  If you are unsure which license is appropriate for your use, please
#  visit www.python-camelot.com or contact project-camelot@conceptive.be
#
#  This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
#  WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
#
#  For use of this library in commercial applications, please contact
#  project-camelot@conceptive.be
#
#  ============================================================================
from PyQt4.QtCore import Qt

from customdelegate import CustomDelegate, DocumentationMetaclass
from camelot.view.controls import editors
from camelot.view.proxy import ValueLoading

class RichTextDelegate(CustomDelegate):
    """Custom delegate for rich text (HTML) string values
  """
  
    __metaclass__ = DocumentationMetaclass
    
    editor = editors.RichTextEditor
    
    def __init__(self, parent=None, editable=True, **kwargs):
        CustomDelegate.__init__(self, parent, editable)
        self.editable = editable
        self._height = self._height * 10
        self._width = self._width * 3
    
    def paint(self, painter, option, index):
        from camelot.view.utils import text_from_richtext
          
        painter.save()
        self.drawBackground(painter, option, index)
        value = unicode(index.model().data(index, Qt.EditRole).toString())

        value_str = u''
        if value not in (None, ValueLoading):
            value_str = ' '.join(text_from_richtext(value))[:256]

        self.paint_text(painter, option, index, value_str)
        painter.restore()        
    





#  ============================================================================
#
#  Copyright (C) 2007-2011 Conceptive Engineering bvba. All rights reserved.
#  www.conceptive.be / project-camelot@conceptive.be
#
#  This file is part of the Camelot Library.
#
#  This file may be used under the terms of the GNU General Public
#  License version 2.0 as published by the Free Software Foundation
#  and appearing in the file license.txt included in the packaging of
#  this file.  Please review this information to ensure GNU
#  General Public Licensing requirements will be met.
#
#  If you are unsure which license is appropriate for your use, please
#  visit www.python-camelot.com or contact project-camelot@conceptive.be
#
#  This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
#  WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
#
#  For use of this library in commercial applications, please contact
#  project-camelot@conceptive.be
#
#  ============================================================================
'''
Created on Jan 18, 2010

@author: tw55413
'''

from progress_page import ProgressPage

class UpdateEntitiesPage(ProgressPage):
    """A progress page that updates each entity in a collection,
    then flushes the entity, and informs all views that the entity
    has been updated.  Subclass this page and implement update_entity
    to make this page do something.
    """
    
    def __init__(self, collection_getter, parent):
        super(UpdateEntitiesPage, self).__init__( parent )
        self._collection_getter = collection_getter
    
    def update_entity(self, entity):
        """Implement this method to update the entities in the 
        collection.
        
        :param entity: the entity that should be updated
        :return: None or a string that will be displayed in the progress
        screen.
        """
        pass
    
    def run(self):
        from sqlalchemy.orm.session import Session
        from camelot.view.remote_signals import get_signal_handler
        signal_handler = get_signal_handler()
        collection = list(self._collection_getter())
        self.update_maximum_signal.emit( len(collection) )
        for i, entity in enumerate(collection):
            message = self.update_entity(entity)
            Session.object_session( entity ).flush( [entity] )
            signal_handler.sendEntityUpdate( self, entity )
            self.update_progress_signal.emit( i, message or '')





#  ============================================================================
#
#  Copyright (C) 2007-2011 Conceptive Engineering bvba. All rights reserved.
#  www.conceptive.be / project-camelot@conceptive.be
#
#  This file is part of the Camelot Library.
#
#  This file may be used under the terms of the GNU General Public
#  License version 2.0 as published by the Free Software Foundation
#  and appearing in the file license.txt included in the packaging of
#  this file.  Please review this information to ensure GNU
#  General Public Licensing requirements will be met.
#
#  If you are unsure which license is appropriate for your use, please
#  visit www.python-camelot.com or contact project-camelot@conceptive.be
#
#  This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
#  WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
#
#  For use of this library in commercial applications, please contact
#  project-camelot@conceptive.be
#
#  ============================================================================

"""
A default Jinja2 environment for the rendering of html in print previews
and others.

The `loader` loads its templates from the camelot/art/templates
folder.  As it is a :class:`jinja2.loaders.ChoiceLoader` object, other
loaders can be appended or prepended to it :attr:`loaders` attribute, to
customize the look of the print previews or reuse the existing style

The `environment` is a :class:`jinja2.environment.Environment` which uses
the `loader` and that can be used with
the :class:`camelot.view.action_steps.print_preview.PrintJinjaTemplate` action
step.
"""

from jinja2.environment import Environment
from jinja2.loaders import ChoiceLoader, PackageLoader

loader = ChoiceLoader( [ PackageLoader( 'camelot.art' ) ] )

class DefaultEnvironment( Environment ):
    
    def __repr__( self ):
        return '<camelot.core.templates.environment>'
    
environment = DefaultEnvironment( loader = loader )



#  ============================================================================
#
#  Copyright (C) 2007-2011 Conceptive Engineering bvba. All rights reserved.
#  www.conceptive.be / project-camelot@conceptive.be
#
#  This file is part of the Camelot Library.
#
#  This file may be used under the terms of the GNU General Public
#  License version 2.0 as published by the Free Software Foundation
#  and appearing in the file license.txt included in the packaging of
#  this file.  Please review this information to ensure GNU
#  General Public Licensing requirements will be met.
#
#  If you are unsure which license is appropriate for your use, please
#  visit www.python-camelot.com or contact project-camelot@conceptive.be
#
#  This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
#  WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
#
#  For use of this library in commercial applications, please contact
#  project-camelot@conceptive.be
#
#  ============================================================================
"""
Logging is an important aspect of every application, as logs provide valuable
feedback to developers and first line support.  Camelot adds some
additional functions on top of the standard Python logging library.

The added functionallity is needed for two reasons :
    
    * Camelot applications are often installed in a distributed fashion.  Thus
      the log files need to be collected to provide meaningfull information to
      the developer.
      
    * Logging should never slow down/freeze the application.  Even when logging
      to files this may happen, since the application never knows for sure if
      a file is really local, and network connections may turn slow at any
      time.

Both issues are resolved by using a logging handler that collects all logs, and
periodically sends them to an http server in the background::
    
    handler = ThreadedHttpHandler('www.example.com:80', '/my_logs/')
    handler.setLevel(logging.INFO)
    logging.root.addHandler(handler)

The logging url could include a part indentifying the user and as such assisting
first line support.
"""

from PyQt4 import QtCore

import getpass
import logging
from logging import handlers
import sys

LOGGER = logging.getLogger('camelot.core.logging')

class ThreadedTimer( QtCore.QThread ):
    """Thread that checks every interval milli seconds if there 
    are logs to be sent to the server"""

    def __init__(self, interval, handler):
        QtCore.QThread.__init__(self)
        self._timer = None
        self._interval = interval
        self._handler = handler

    def run(self):
        self._timer = QtCore.QTimer()
        self._timer.timeout.connect( self._handler.timeout )
        self._timer.start(self._interval)
        self.exec_()

class ThreadedHttpHandler(handlers.HTTPHandler):
    """An Http Logging handler that does the logging itself in a different 
    thread, to prevent slow down of the main thread"""

    def __init__(self, host, url, method='GET'):
        handlers.HTTPHandler.__init__(self, host, url, method)
        self._records_to_emit = []
        self._threaded_timer = ThreadedTimer(1000, self)
        self._threaded_timer.start()

    @QtCore.pyqtSlot()
    def timeout(self):
        while len(self._records_to_emit):
            record = self._records_to_emit.pop()
            handlers.HTTPHandler.emit(self, record)

    def emit(self, record):
        self._records_to_emit.append(record)

class ThreadedAwsHandler(logging.Handler):
    """A logging handler that sends the logs to an AWS queue through the
    SQS, this handler requires the boto library"""
    
    def __init__(self, access_key, secret_access_key, queue_name, 
                 revision=0, connection_kwargs={}):
        """:param connection_kwargs: arguments to be used when creating the
        underlying boto connection to aws"""
        logging.Handler.__init__(self)
        self._access_key = access_key
        self._secret_access_key = secret_access_key
        self._connection_kwargs = connection_kwargs
        self._queue_name = queue_name
        self._records_to_emit = []
        self._queue = None
        self._connected = True
        self._threaded_timer = ThreadedTimer(1000, self)
        self._threaded_timer.start()
        try:
            if sys.platform.startswith('win'):
                # on windows getuser() uses the USERNAME env var
                # from sys.getfilesystemencoding documentation:
                # On Windows NT+, file names are Unicode natively, so no conversion is performed. 
                # getfilesystemencoding() still returns 'mbcs', as this is the encoding that applications 
                # should use when they explicitly want to convert Unicode strings 
                # to byte strings that are equivalent when used as file names.
                self._user = getpass.getuser().decode('mbcs')
            else:
                self._user = getpass.getuser()
        except Exception:
            self._user = getpass.getuser().encode('ascii', 'ignore')
        self._revision = revision
        
    def emit(self, record):
        # inspired by the code in logging.handlers.SocketHandler
        import json
        ei = record.exc_info
        #
        # prevent infinite loops when the boto lib logs something when
        # a log is sent to the queue
        #
        if record.name and record.name.startswith('boto'):
            return
        if ei:
            self.format(record)
            record.exc_info = None
        record_dict = dict( user=self._user, revision=self._revision )
        record_dict.update( record.__dict__ )
        self._records_to_emit.append( json.dumps( record_dict ) )
        if ei:
            record.exc_info = ei  # for next handler
        
    @QtCore.pyqtSlot()
    def timeout(self):
        from boto.sqs.message import Message
        if not self._queue and self._connected:
            try:
                from boto.sqs.connection import SQSConnection
                sqs_connection = SQSConnection(self._access_key, 
                                               self._secret_access_key,
                                               **self._connection_kwargs)
                self._queue = sqs_connection.get_queue( self._queue_name )
            except Exception, e:
                LOGGER.error('Could not connect to logging queue %s'%self._queue_name, exc_info=e)
                self._connected = False
        while len(self._records_to_emit) and self._connected:
            record = self._records_to_emit.pop()
            self._queue.write( Message(body=record) )

class CloudLaunchHandler(ThreadedAwsHandler):
    """A logging handler that sends the logs to the Cloud Launch service,
    requires the cloudlaunch library, and only works with applications tested
    or deployed through cloudlaunch
    
    The cloudlaunch record can be obtained using the get_cloud_record
    method of the cloudlaunch.resources module.
    """
    
    def __init__(self, cloud_record, connection_kwargs={}):
        """:param cloud_record: the cloud record describing the application
        """
        queue_name = 'cloudlaunch-%s-%s-logging'%(cloud_record.author.replace(' ','_'), cloud_record.name.replace(' ','_') )
        ThreadedAwsHandler.__init__(self, 
                                    cloud_record.public_access_key, 
                                    cloud_record.public_secret_key, 
                                    queue_name,
                                    revision = cloud_record.revision,
                                    connection_kwargs = connection_kwargs)



#  ============================================================================
#
#  Copyright (C) 2007-2011 Conceptive Engineering bvba. All rights reserved.
#  www.conceptive.be / project-camelot@conceptive.be
#
#  This file is part of the Camelot Library.
#
#  This file may be used under the terms of the GNU General Public
#  License version 2.0 as published by the Free Software Foundation
#  and appearing in the file license.txt included in the packaging of
#  this file.  Please review this information to ensure GNU
#  General Public Licensing requirements will be met.
#
#  If you are unsure which license is appropriate for your use, please
#  visit www.python-camelot.com or contact project-camelot@conceptive.be
#
#  This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
#  WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
#
#  For use of this library in commercial applications, please contact
#  project-camelot@conceptive.be
#
#  ============================================================================

from PyQt4 import QtGui, QtCore
  
from camelot.admin.action import ActionStep
from camelot.core.templates import environment

class OpenFile( ActionStep ):
    """
    Open a file with the preferred application from the user.  The absolute
    path is preferred, as this is most likely to work when running from an
    egg and in all kinds of setups.
    
    :param path: the absolute path to the file to open
    
    The :keyword:`yield` statement will return :keyword:`True` if the file was
    opend successfull.
    """
        
    def __init__( self, path ):
        self._path = path

    def get_path( self ):
        """
        :return: the path to the file that will be opened, use this method
        to verify the content of the file in unit tests
        """
        return self._path

    @classmethod
    def create_temporary_file( self, suffix ):
        """
        Create a temporary filename that can be used to write to, and open
        later on.
        
        :param suffix: the suffix of the file to create
        :return: the filename of the temporary file
        """
        import tempfile
        import os
        file_descriptor, file_name = tempfile.mkstemp( suffix=suffix )
        os.close( file_descriptor )
        return file_name
        
    def gui_run( self, gui_context ):
        #
        # support for windows shares
        #
        if not self._path.startswith(r'\\'):
            url = QtCore.QUrl.fromLocalFile( self._path )
        else:
            url = QtCore.QUrl( self._path, QtCore.QUrl.TolerantMode )
        return QtGui.QDesktopServices.openUrl( url )
    
class OpenStream( OpenFile ):
    """Write a stream to a temporary file and open that file with the 
    preferred application of the user.
    
    :param stream: the stream to write to a file
    :param suffix: the suffix of the temporary file
    """

    def __init__( self, stream, suffix='.txt' ):
        import os
        import tempfile
        file_descriptor, file_name = tempfile.mkstemp( suffix=suffix )
        output_stream = os.fdopen( file_descriptor, 'wb' )
        output_stream.write( stream.read() )
        output_stream.close()
        super( OpenStream, self ).__init__( file_name )

class OpenString( OpenFile ):
    
    def __init__( self, string, suffix='.txt' ):
        """Write a string to a temporary file and open that file with the
        preferred application of the user.
        
        :param string: the string to write to a file
        :param suffix: the suffix of the temporary file
        """
        import os
        import tempfile
        file_descriptor, file_name = tempfile.mkstemp( suffix=suffix )
        output_stream = os.fdopen( file_descriptor, 'wb' )
        output_stream.write( string )
        output_stream.close()
        super( OpenString, self ).__init__( file_name )
        
class OpenJinjaTemplate( OpenStream ):
    """Render a jinja template into a temporary file and open that
    file with the prefered application of the user.
    
    :param environment: a :class:`jinja2.Environment` object to be used
        to load templates from.
        
    :param template: the name of the template as it can be fetched from
        the Jinja environment.
    
    :param suffix: the suffix of the temporary file to create, this will
        determine the application used to open the file.
        
    :param context: a dictionary with objects to be used when rendering
        the template
    """
    
    def __init__( self,
                  template, 
                  context={},
                  environment = environment,
                  suffix='.txt' ):
        from cStringIO import StringIO
        template = environment.get_template( template )
        template_stream = template.stream( context )
        output_stream = StringIO()
        template_stream.dump( output_stream, encoding='utf-8' )
        output_stream.seek( 0 )
        super( OpenJinjaTemplate, self).__init__( output_stream, suffix=suffix )

class WordJinjaTemplate( OpenFile ):
    """Render a jinja template into a temporary file and open that
    file with microsoft word through the use of COM objects.
    
    :param environment: a :class:`jinja2.Environment` object to be used
        to load templates from.
        
    :param template: the name of the template as it can be fetched from
        the Jinja environment.
    
    :param suffix: the suffix of the temporary file to create, this will
        determine the application used to open the file.
        
    :param context: a dictionary with objects to be used when rendering
        the template
    """
   
    def __init__( self,
                  template, 
                  context={},
                  environment = environment,
                  suffix='.xml' ):
        self.file_name = self.create_temporary_file( suffix )
        template = environment.get_template( template )
        template_stream = template.stream( context )
        template_stream.dump( open( self.file_name, 'wb' ), encoding='utf-8' )
        
    def gui_run( self, gui_context ):
        import pythoncom
        import win32com.client
        pythoncom.CoInitialize()
        word_app = win32com.client.Dispatch("Word.Application")
        word_app.Visible = True
        doc = word_app.Documents.Open( self.file_name )
        doc.Activate()
        word_app.Activate()



#  ============================================================================
#
#  Copyright (C) 2007-2011 Conceptive Engineering bvba. All rights reserved.
#  www.conceptive.be / project-camelot@conceptive.be
#
#  This file is part of the Camelot Library.
#
#  This file may be used under the terms of the GNU General Public
#  License version 2.0 as published by the Free Software Foundation
#  and appearing in the file license.txt included in the packaging of
#  this file.  Please review this information to ensure GNU
#  General Public Licensing requirements will be met.
#
#  If you are unsure which license is appropriate for your use, please
#  visit www.python-camelot.com or contact project-camelot@conceptive.be
#
#  This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
#  WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
#
#  For use of this library in commercial applications, please contact
#  project-camelot@conceptive.be
#
#  ============================================================================

import os
import logging
logger = logging.getLogger('camelot.admin.application_admin')

from PyQt4 import QtCore, QtGui

from camelot.view.model_thread import model_function
from camelot.core.backup import BackupMechanism
from camelot.core.utils import ugettext_lazy as _
from camelot.view import art
from camelot.view import database_selection

_application_admin_ = []

#
# The translations data needs to be kept alive during the
# running of the application
#
_translations_data_ = []

def get_application_admin():
    if not len(_application_admin_):
        raise Exception('No application admin class has been constructed yet')
    return _application_admin_[0]

class ApplicationAdmin(QtCore.QObject):
    """The ApplicationAdmin class defines how the application should look
like, it also ties Python classes to their associated 
:class:`camelot.admin.object_admin.ObjectAdmin` class or subclass.  It's
behaviour can be steered by overwriting its static attributes or it's
methods :

.. attribute:: name

    The name of the application, as it will appear in the title of the main
    window.

.. attribute:: application_url

    The url of the web site where the user can find more information on
    the application.

.. attribute:: help_url

    Points to either a local html file or a web site that contains the
    documentation of the application.

.. attribute:: author

    The name of the author of the application
    
.. attribute:: domain

    The domain name of the author of the application, eg 'mydomain.com', this
    domain will be used to store settings of the application.
    
.. attribute:: version
    
    A string with the version of the application

.. attribute:: backup_mechanism

    A subclass of :class:`camelot.core.backup.BackupMechanism` that enables 
    the application to perform backups an restores.
    
.. attribute:: database_profile_wizard
    
    The wizard that should be used to create new database profiles. Defaults
    to :class:`camelot.view.database_selection.ProfileWizard`
    
.. attribute:: database_selection

    if this is set to True, present the user with a database selection
    wizard prior to starting the application.  Defaults to :keyword:`False`.
    """

    backup_mechanism = BackupMechanism
    database_profile_wizard = database_selection.ProfileWizard

    name = 'Camelot'
    application_url = None
    help_url = 'http://www.python-camelot.com/docs.html'
    author = 'Conceptive Engineering'
    domain = 'python-camelot.com'

    version = '1.0'
    admins = {}

    # This signal is emitted whenever the sections are changed, and the views
    # should be updated
    sections_changed_signal = QtCore.pyqtSignal()
    # This signal is emitted whenever the tile of the main window needs to
    # be changed.
    title_changed_signal = QtCore.pyqtSignal(str)
    # Emitted whenever the application actions need to be changed
    actions_changed_signal = QtCore.pyqtSignal()

    database_selection = False

    def __init__(self):
        """Construct an ApplicationAdmin object and register it as the 
        prefered ApplicationAdmin to use througout the application"""
        QtCore.QObject.__init__(self)
        _application_admin_.append(self)
        #
        # Cache created ObjectAdmin objects
        #
        self._object_admin_cache = {}

    def register(self, entity, admin_class):
        """Associate a certain ObjectAdmin class with another class.  This
        ObjectAdmin will be used as default to render object the specified
        type.
        
        :param entity: :class:`class`
        :param admin_class: a subclass of 
            :class:`camelot.admin.object_admin.ObjectAdmin` or
            :class:`camelot.admin.entity_admin.EntityAdmin`
        """
        self.admins[entity] = admin_class

    @model_function
    def get_sections(self):
        """A list of :class:`camelot.admin.section.Section` objects,
        these are the sections to be displayed in the left panel.
        
        .. image:: /_static/picture2.png
        """
        from camelot.admin.section import Section
        
        return [ Section( _('Relations'), self ),
                 Section( _('Configuration'), self ),
                 ]
        
    def get_related_admin(self, cls):
        """Get the default :class:`camelot.admin.object_admin.ObjectAdmin` class
        for a specific class, return None, if not known.  The ObjectAdmin
        should either be registered through the :meth:`register` method or be
        defined as an inner class with name :keyword:`Admin` of the entity.

        :param entity: a :class:`class`
        
        """
        return self.get_entity_admin( cls )
    
    def get_entity_admin(self, entity):
        """Get the default :class:`camelot.admin.object_admin.ObjectAdmin` class
        for a specific entity, return None, if not known.  The ObjectAdmin
        should either be registered through the :meth:`register` method or be
        defined as an inner class with name :keyword:`Admin` of the entity.

        :param entity: a :class:`class`
        
        deprecated : use get_related_admin instead
        """

        admin_class = None
        try:
            admin_class = self.admins[entity]
        except KeyError:
            pass
        if not admin_class and hasattr(entity, 'Admin'):
            admin_class = entity.Admin
        if admin_class:
            try:
                return self._object_admin_cache[admin_class]
            except KeyError:
                admin = admin_class(self, entity)
                self._object_admin_cache[admin_class] = admin
                return admin

    def create_main_window(self):
        """Create the main window that will be shown when the application
        starts up.  By default, returns an instance of 
         :class:`camelot.view.mainwindow.MainWindow`
         
        :return: a :class:`PyQt4.QtGui.QWidget`
        """
        from camelot.view.mainwindow import MainWindow
        mainwindow = MainWindow(self)
        shortcut_versions = QtGui.QShortcut(
            QtGui.QKeySequence( QtCore.Qt.CTRL+QtCore.Qt.ALT+QtCore.Qt.Key_V ),
            mainwindow
        )
        shortcut_versions.activated.connect( self.show_versions )
        shortcut_dump_state = QtGui.QShortcut(
            QtGui.QKeySequence( QtCore.Qt.CTRL+QtCore.Qt.ALT+QtCore.Qt.Key_D ),
            mainwindow
        )
        shortcut_dump_state.activated.connect( self.dump_state )
        shortcut_read_null = QtGui.QShortcut(
            QtGui.QKeySequence( QtCore.Qt.CTRL+QtCore.Qt.ALT+QtCore.Qt.Key_0 ),
            mainwindow
        )
        shortcut_read_null.activated.connect( self.read_null )
        return mainwindow

    @QtCore.pyqtSlot()
    def show_versions(self):
        """Pops up a messagebox showing the version of certain
        libraries used.  This is for debugging purposes."""
        logger.debug('showing about message box with versions')
        abtmsg = self.get_versions()
        QtGui.QMessageBox.about(None, 'Versions', abtmsg)
        logger.debug('about message with versions closed')

    def get_actions(self):
        """
        :return: a list of :class:`camelot.admin.application_action.ApplicationAction` objects
            that should be added to the menu and the icon bar for this application
        """
        return []

    def get_name(self):
        """
        :return: the name of the application, by default this is the class
            attribute name"""
        return unicode( self.name )

    def get_version(self):
        """:return: string representing version of the application, by default this
                    is the class attribute verion"""
        return self.version

    def get_icon(self):
        """:return: the :class:`camelot.view.art.Icon` that should be used for the application"""
        from camelot.view.art import Icon
        return Icon('tango/32x32/apps/system-users.png').getQIcon()

    def get_splashscreen(self):
        """:return: a :class:`PyQt4.QtGui.QPixmap` to be used as splash screen"""
        from camelot.view.art import Pixmap
        return Pixmap('splashscreen.png').getQPixmap()

    def get_organization_name(self):
        return self.author

    def get_organization_domain(self):
        return self.domain

    def get_help_url(self):
        """:return: a :class:`PyQt4.QtCore.QUrl` pointing to the index page for help"""
        from PyQt4.QtCore import QUrl
        if self.help_url:
            return QUrl( self.help_url )

    def get_whats_new(self):
        """:return: a widget that has a show() method """
        return None

    def get_affiliated_url(self):
        """:return: a :class:`PyQt4.QtCore.QUrl` pointing to an affiliated webpage

        When this method returns a QUrl, an additional item will be available
        in the 'Help' menu, when clicked the system browser will be opened
        an pointing to this url.

        This can be used to connect the user to a website that is used a lot
        in the organization, but hard to remember.
        """
        from PyQt4.QtCore import QUrl
        if self.application_url:
            return QUrl( self.application_url )

    def get_remote_support_url(self):
        """:return: a :class:`PyQt4.QtCore.QUrl` pointing to a page to get remote support

        When this method returns a QUrl, an additional item will be available
        in the 'Help' menu, when clicked the system browser will be opened
        an pointing to this url.

        This can be used to connect the user to services like logmein.com, an
        online ticketing system or others.
        """
        return None

    def get_stylesheet(self):
        """
        :return: a string with the qt stylesheet to be used for this application as a string
        or None if no stylesheet needed.

        Camelot comes with a couple of default stylesheets :

         * stylesheet/office2007_blue.qss
         * stylesheet/office2007_black.qss
         * stylesheet/office2007_silver.qss

        Have a look at the default implementation to use another stylesheet.
        """
        return art.read('stylesheet/office2007_blue.qss')

    def _load_translator_from_file( self, 
                                    module_name, 
                                    file_name, 
                                    directory = '', 
                                    search_delimiters = '_', 
                                    suffix = '.qm' ):
        """
        Tries to create a translator based on a file stored within a module.
        The file is loaded through the pkg_resources, to enable loading it from
        within a Python egg.  This method tries to mimic the behavior of
        :meth:`QtCore.QTranslator.load` while looking for an appropriate
        translation file.
        
        :param module_name: the name of the module in which to look for
            the translation file with pkg_resources.
        :param file_name: the filename of the the tranlations file, without 
            suffix
        :param directory: the directory, relative to the module in which
            to look for translation files
        :param suffix: the suffix of the filename
        :return: :keyword:None if unable to load the file, otherwise a
            :obj:`QtCore.QTranslator` object.
        """
        from camelot.core.resources import resource_string
        file_names = []
        for search_delimiter in search_delimiters:
            file_name_parts = file_name.split( search_delimiter )
            for i in range( len(file_name_parts) ):
                partial_file_name = search_delimiter.join( file_name_parts[:len(file_name_parts)-i] )
                file_names.append( partial_file_name + suffix )
                file_names.append( partial_file_name )
        translations = None
        for file_name in file_names:
            try:
                translations = resource_string( module_name, os.path.join(directory,file_name) )
                break
            except IOError:
                pass
        if translations:
            _translations_data_.append( translations ) # keep the data alive
            translator = QtCore.QTranslator()
            # PySide workaround for missing loadFromData method
            if not hasattr( translator, 'loadFromData' ):
                return
            if translator.loadFromData( translations ):
                logger.info("add translation %s" % (directory + file_name))
                return translator
        
    def get_translator(self):
        """Reimplement this method to add application specific translations
        to your application.  The default method returns a list with the
        default Qt and the default Camelot translator for the current system
        locale.  Call :meth:`QLocale.setDefault` before this method is called
        if you want to load different translations then the system default.

        :return: a list of :obj:`QtCore.QTranslator` objects that should be 
            used to translate the application
        """
        translators = []
        qt_translator = QtCore.QTranslator()
        locale_name = QtCore.QLocale().name()
        language_name = locale_name.split('_')[0]
        logger.info( u'using locale %s'%locale_name )
        if qt_translator.load( "qt_" + locale_name,
                              QtCore.QLibraryInfo.location( QtCore.QLibraryInfo.TranslationsPath ) ):
            translators.append( qt_translator )
        camelot_translator = self._load_translator_from_file( 'camelot', 
                                                              'camelot',
                                                              'art/translations/%s/LC_MESSAGES/'%language_name )
        if camelot_translator:
            translators.append( camelot_translator )
        else:
            logger.debug( 'no camelot translations found for %s'%locale_name )
        return translators

    def get_about(self):
        """:return: the content of the About dialog, a string with html
                    syntax"""
        import datetime
        from camelot.core import license
        today = datetime.date.today()
        return """<b>Camelot</b><br/>
                  Building desktop applications at warp speed
                  <p>
                  Copyright &copy; 2007-%s Conceptive Engineering.
                  All rights reserved.
                  </p>
                  <p>
                  %s
                  </p>
                  <p>
                  http://www.python-camelot.com<br/>
                  http://www.conceptive.be
                  </p>
                  """%(today.year, license.license_type)

    def get_versions(self):
        """
        :return: html which displays the versions of used libs for development
        """
        import sys
        import sqlalchemy
        import elixir
        import chardet
        import jinja2
        import xlrd
        import xlwt
                
        return """<em>Python:</em> <b>%s</b><br>
                  <em>Qt:</em> <b>%s</b><br>
                  <em>PyQt:</em> <b>%s</b><br>
                  <em>SQLAlchemy:</em> <b>%s</b><br>
                  <em>Elixir:</em> <b>%s</b><br>
                  <em>Chardet:</em> <b>%s</b><br>
                  <em>Jinja:</em> <b>%s</b><br>
                  <em>xlrd:</em> <b>%s</b><br>
                  <em>xlwt:</em> <b>%s</b><br><br>
                  <em>path:<br></em> %s""" % ('.'.join([str(el) for el in sys.version_info]),
                                              float('.'.join(str(QtCore.QT_VERSION_STR).split('.')[0:2])),
                                              QtCore.PYQT_VERSION_STR,
                                              sqlalchemy.__version__,
                                              elixir.__version__,
                                              chardet.__version__,
                                              jinja2.__version__,
                                              xlrd.__VERSION__,
                                              xlwt.__VERSION__,
                                              unicode(sys.path))
    
    def read_null(self):
        """Create a segmentation fault by reading null, this is to test
        the faulthandling functions.  this method is triggered by pressing
        :kbd:`Ctrl-Alt-0` in the GUI"""
        ok = QtGui.QMessageBox.critical( None, 
                                         'Experimental segfault',
                                         'Are you sure you want to segfault the application',
                                         buttons = QtGui.QMessageBox.No | QtGui.QMessageBox.Yes )
        if ok == QtGui.QMessageBox.Yes:
            import faulthandler
            faulthandler._read_null()
    
    def dump_state(self):
        """Dump the state of the application to the output, this method is
        triggered by pressing :kbd:`Ctrl-Alt-D` in the GUI"""
        from camelot.view.model_thread import post
        from camelot.view.register import dump_register
        from camelot.view.proxy.collection_proxy import CollectionProxy

        import gc
        gc.collect()

            
        dump_register()
        
        def dump_session_state():
            import collections
            
            from camelot.model.authentication import Person

            print '======= begin session =============='
            type_counter = collections.defaultdict(int)
            for o in Person.query.session:
                type_counter[type(o).__name__] += 1
            for k,v in type_counter.items():
                print k,v
            print '====== end session =============='

        post( dump_session_state )

        for o in gc.get_objects():
            if isinstance(o, CollectionProxy):
                print o
                for r in gc.get_referrers(o):
                    print ' ', type(r).__name__
                    for rr in gc.get_referrers(r):
                        print  '  ', type(rr).__name__
                
    def get_default_field_attributes(self, type_, field):
        """Returns the default field attributes"""
        from camelot.core.view.field_attributes import \
            _sqlalchemy_to_python_type_
        return _sqlalchemy_to_python_type_[type_](field)

    def backup(self, main_window):
        from camelot.view.wizard.backup import BackupWizard
        wizard = BackupWizard(self.backup_mechanism, main_window)
        wizard.exec_()

    def restore(self, main_window):
        from camelot.view.wizard.backup import RestoreWizard
        wizard = RestoreWizard(self.backup_mechanism, main_window)
        wizard.exec_()



#  ============================================================================
#
#  Copyright (C) 2007-2011 Conceptive Engineering bvba. All rights reserved.
#  www.conceptive.be / project-camelot@conceptive.be
#
#  This file is part of the Camelot Library.
#
#  This file may be used under the terms of the GNU General Public
#  License version 2.0 as published by the Free Software Foundation
#  and appearing in the file license.txt included in the packaging of
#  this file.  Please review this information to ensure GNU
#  General Public Licensing requirements will be met.
#
#  If you are unsure which license is appropriate for your use, please
#  visit www.python-camelot.com or contact project-camelot@conceptive.be
#
#  This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
#  WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
#
#  For use of this library in commercial applications, please contact
#  project-camelot@conceptive.be
#
#  ============================================================================

"""
Helper functions to search through a collection of entities
"""
import logging

LOGGER = logging.getLogger('camelot.view.search')

import sqlalchemy.types
from sqlalchemy.sql import operators

import camelot.types

def create_entity_search_query_decorator(admin, text):
    """create a query decorator to search through a collection of entities
    @param admin: the admin interface of the entity
    @param text: the text to search for
    @return: a function that can be applied to a query to make the query filter
    only the objects related to the requested text or None if no such decorator
    could be build
    """
    from elixir import entities
    from sqlalchemy import orm, sql
    from camelot.view import utils

    if len(text.strip()):
        from sqlalchemy import Unicode, or_
        # arguments for the where clause
        args = []
        # join conditions : list of join entities
        joins = []

        def append_column(c):
            """add column c to the where clause using a clause that
            is relevant for that type of column"""
            arg = None
            if issubclass(c.type.__class__, camelot.types.Color):
                pass
            elif issubclass(c.type.__class__, camelot.types.File):
                pass
            elif issubclass(c.type.__class__, camelot.types.Code):
                codes = [u'%%%s%%'%s for s in text.split(c.type.separator)]
                codes = codes + ['%']*(len(c.type.parts) - len(codes))
                arg = c.like( codes )
            elif issubclass(c.type.__class__, camelot.types.VirtualAddress):
                arg = c.like(('%', '%'+text+'%'))
            elif issubclass(c.type.__class__, camelot.types.Image):
                pass
            elif issubclass(c.type.__class__, sqlalchemy.types.Integer):
                try:
                    arg = (c==utils.int_from_string(text))
                except Exception, utils.ParsingError:
                    pass
            elif issubclass(c.type.__class__, sqlalchemy.types.Date):
                try:
                    arg = (c==utils.date_from_string(text))
                except Exception, utils.ParsingError:
                    pass
            elif issubclass(c.type.__class__, sqlalchemy.types.Float):
                try:
                    float_value = utils.float_from_string(text)
                    precision = c.type.precision
                    if isinstance(precision, (tuple)):
                        precision = precision[1]
                    delta = 0.1**precision
                    arg = sql.and_(c>=float_value-delta, c<=float_value+delta)
                except Exception, utils.ParsingError:
                    pass
            elif issubclass(c.type.__class__, (Unicode, )) or \
                            (hasattr(c.type, 'impl') and \
                             issubclass(c.type.impl.__class__, (Unicode, ))):
                LOGGER.debug('look in column : %s'%c.name)
                arg = operators.ilike_op(c, '%'+text+'%')

            if arg is not None:
                arg = sql.and_(c != None, arg)
                args.append(arg)

        if admin.search_all_fields:
            search_tables = set([admin.entity.table])
            for entity in entities:
                if issubclass(admin.entity, entity):
                    search_tables.add(entity.table)
            for table in search_tables:
                for column in table._columns:
                    append_column(column)

        for column_name in admin.list_search:
            path = column_name.split('.')
            target = admin.entity
            for path_segment in path:
                mapper = orm.class_mapper(target)
                property = mapper.get_property(path_segment, resolve_synonyms=True)
                if isinstance(property, orm.properties.PropertyLoader):
                    joins.append(getattr(target, path_segment))
                    target = property._get_target().class_
                else:
                    append_column(property.columns[0])
                    #args.append(property.columns[0].like('%'+text+'%'))

        def create_query_decorator(joins, args):
            """Bind the join and args to a query decorator function"""

            def query_decorator(query):
                """The actual query decorator, call this function with a query
                as its first argument and it will return a query with a where
                clause for searching the resultset of the original query"""
                for join in joins:
                    query = query.outerjoin(join)
                if len(args):
                    if len(args)>1:
                        query = query.filter(or_(*args))
                    else:
                        query = query.filter(args[0])
                return query

            return query_decorator

        return create_query_decorator(joins, args)





#  ============================================================================
#
#  Copyright (C) 2007-2011 Conceptive Engineering bvba. All rights reserved.
#  www.conceptive.be / project-camelot@conceptive.be
#
#  This file is part of the Camelot Library.
#
#  This file may be used under the terms of the GNU General Public
#  License version 2.0 as published by the Free Software Foundation
#  and appearing in the file license.txt included in the packaging of
#  this file.  Please review this information to ensure GNU
#  General Public Licensing requirements will be met.
#
#  If you are unsure which license is appropriate for your use, please
#  visit www.python-camelot.com or contact project-camelot@conceptive.be
#
#  This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
#  WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
#
#  For use of this library in commercial applications, please contact
#  project-camelot@conceptive.be
#
#  ============================================================================

"""Module for managing imports"""

import csv
import codecs

from PyQt4 import QtGui
from PyQt4 import QtCore

from camelot.core.utils import xls2list
from camelot.core.utils import ugettext as _
from camelot.core.utils import ugettext_lazy

from camelot.view.art import Pixmap, ColorScheme
from camelot.view.model_thread import post
from camelot.view.wizard.pages.select import SelectFilePage
from camelot.view.wizard.pages.progress_page import ProgressPage
from camelot.view.controls.editors.one2manyeditor import One2ManyEditor
from camelot.view.proxy.collection_proxy import CollectionProxy

import logging
logger = logging.getLogger('camelot.view.wizard.importwizard')

class RowData(object):
    """Class representing the data in a single row of the imported file as an
    object with attributes column_1, column_2, ..., each representing the data
    in a single column of that row.

    since the imported file might contain less columns than expected, the
    RowData object returns None for not existing attributes"""

    def __init__(self, row_number, row_data):
        """:param row_data: a list containing the data
        [column_1_data, column_2_data, ...] for a single row
        """
        self.id = row_number + 1
        for i, data in enumerate(row_data):
            self.__setattr__('column_%i' % i, data)

    def __getattr__(self, attr_name):
        return None


# see http://docs.python.org/library/csv.html
class UTF8Recoder:
    """Iterator that reads an encoded stream and reencodes the input to
    UTF-8."""

    def __init__(self, f, encoding):
        self.reader = codecs.getreader(encoding)(f)

    def __iter__(self):
        return self

    def next(self):
        return self.reader.next().encode('utf-8')


# see http://docs.python.org/library/csv.html
class UnicodeReader:
    """A CSV reader which will iterate over lines in the CSV file "f", which is
    encoded in the given encoding."""

    def __init__(self, f, dialect=csv.excel, encoding='utf-8', **kwds):
        f = UTF8Recoder(f, encoding)
        self.reader = csv.reader(f, dialect=dialect, **kwds)

    def next(self):
        row = self.reader.next()
        return [unicode(s, 'utf-8') for s in row]

    def __iter__(self):
        return self


class CsvCollectionGetter(object):
    """Returns data from csv file as a list of RowData objects"""

    def __init__(self, filename):
        self.filename = filename
        self._data = None

    def __call__(self):
        if self._data==None:
            self._data = []
            import chardet

            detected = chardet.detect(open(self.filename).read())['encoding']
            enc = detected or 'utf-8'

            items = UnicodeReader(open(self.filename), encoding=enc)

            self._data = [
                RowData(i, row_data)
                for i, row_data in enumerate(items)
            ]

        return self._data


class XlsCollectionGetter(object):
    """Returns the data from excel file as a list of RowData objects"""

    def __init__(self, filename, encoding='utf-8'):
        self.filename = filename
        self.encoding = encoding
        self._data = None

    def __call__(self):
        if self._data == None:
            rows = xls2list(self.filename)[1:] # we skip the first row :)

            self._data = [
                RowData(i, row_data)
                for i, row_data in enumerate(rows)
            ]

        return self._data


class RowDataAdminDecorator(object):
    """Decorator that transforms the Admin of the class to be imported to an
    Admin of the RowData objects to be used when previewing and validating the
    data to be imported.

    based on the field attributes of the original mode, it will turn the
    background color pink if the data is invalid for being imported.
    """

    def __init__(self, object_admin):
        """:param object_admin: the object_admin object that will be
        decorated"""
        self._object_admin = object_admin
        self._new_field_attributes = {}
        self._columns = None

    def __getattr__(self, attr):
        return getattr(self._object_admin, attr)

    def create_validator(self, model):
        """Creates a validator that validates the data to be imported, the
        validator will check if the background of the cell is pink, and if it
        is it will mark that object as invalid.
        """
        from camelot.admin.validator.object_validator import ObjectValidator

        class NewObjectValidator(ObjectValidator):

            def objectValidity(self, obj):
                columns = self.admin.get_columns()
                dynamic_attributes = self.admin.get_dynamic_field_attributes(
                    obj,
                    [c[0] for c in columns]
                )
                for attrs in dynamic_attributes:
                    if attrs['background_color'] == ColorScheme.pink_1:
                        logger.debug('we have an invalid field')
                        return ['invalid field']
                return []

        return NewObjectValidator(self, model)

    def get_fields(self):
        return self.get_columns()

    def flush(self, obj):
        """When flush is called, don't do anything, since we'll only save the
        object when importing them for real"""
        pass
    
    def delete(self, obj):
        pass

    def get_field_attributes(self, field_name):
        return self._new_field_attributes[field_name]

    def get_static_field_attributes(self, field_names):
        for _field_name in field_names:
            yield {'editable':True}

    def get_dynamic_field_attributes(self, obj, field_names):
        for field_name in field_names:
            attributes = self.get_field_attributes(field_name)
            string_value = attributes['getter'](obj)
            valid = True
            value = None
            if 'from_string' in attributes:
                try:
                    value = attributes['from_string'](string_value)
                except Exception:
                    valid = False
                # 0 is valid
                if value != 0 and not value and not attributes['nullable']:
                    valid = False
            if valid:
                yield {'background_color':None}
            else:
                yield {'background_color':ColorScheme.pink_1}

    def new_field_attributes(self, i, original_field_attributes, original_field):
        from camelot.view.controls import delegates

        def create_getter(i):
            return lambda o:getattr(o, 'column_%i'%i)

        attributes = dict(original_field_attributes)
        attributes['delegate'] = delegates.PlainTextDelegate
        attributes['python_type'] = str
        attributes['original_field'] = original_field
        attributes['getter'] = create_getter(i)

        # remove some attributes that might disturb the import wizard
        for attribute in ['background_color', 'tooltip']:
            attributes[attribute] = None

        self._new_field_attributes['column_%i' %i] = attributes

        return attributes

    def get_columns(self):
        if self._columns:
            return self._columns

        original_columns = self._object_admin.get_columns()
        new_columns = [
            (
                'column_%i' %i,
                self.new_field_attributes(i, attributes, original_field)
            )
            for i, (original_field, attributes) in enumerate(original_columns)
            if attributes.get('editable',  True)
        ]

        self._columns = new_columns

        return new_columns


class DataPreviewPage(QtGui.QWizardPage):
    """DataPreviewPage is the previewing page for the import wizard"""

    def __init__(self, parent=None, model=None, collection_getter=None):
        from camelot.view.controls.editors import NoteEditor
        super(DataPreviewPage, self).__init__(parent)
        assert model
        assert collection_getter
        self.setTitle(_('Data Preview'))
        self.setSubTitle(_('Please review the data below.'))
        self._complete = False
        self.model = model
        validator = self.model.get_validator()
        validator.validity_changed_signal.connect(self.update_complete)
        model.layoutChanged.connect(self.validate_all_rows)
        post(validator.validate_all_rows)
        self.collection_getter = collection_getter

        icon = 'tango/32x32/mimetypes/x-office-spreadsheet.png'
        self.setPixmap(QtGui.QWizard.LogoPixmap, Pixmap(icon).getQPixmap())

        self.previewtable = One2ManyEditor(
            admin = model.get_admin(),
            parent = self,
            create_inline = True,
            vertical_header_clickable = False,
        )
        self._note = NoteEditor()
        self._note.set_value(None)

        ly = QtGui.QVBoxLayout()
        ly.addWidget(self.previewtable)
        ly.addWidget(self._note)
        self.setLayout(ly)

        self.setCommitPage(True)
        self.setButtonText(QtGui.QWizard.CommitButton, _('Import'))
        self.update_complete()

    @QtCore.pyqtSlot()
    def validate_all_rows(self):
        validator = self.model.get_validator()
        post(validator.validate_all_rows, self._all_rows_validated)

    def _all_rows_validated(self, *args):
        self.update_complete(0)

    @QtCore.pyqtSlot(int)
    def update_complete(self, row=0):
        self._complete = (self.model.get_validator().number_of_invalid_rows()==0)
        self.completeChanged.emit()
        if self._complete:
            self._note.set_value(None)
        else:
            self._note.set_value(_(
                'Please correct the data above before proceeding with the '
                'import.<br/>Incorrect cells have a pink background.'
            ))

    def initializePage(self):
        """Gets all info needed from SelectFilePage and feeds table"""
        filename = self.field('datasource').toString()
        self._complete = False
        self.completeChanged.emit()
        self.model.set_collection_getter(self.collection_getter(filename))
        self.previewtable.set_value(self.model)
        self.validate_all_rows()

    def validatePage(self):
        answer = QtGui.QMessageBox.question(
            self,
            _('Proceed with import'),
            _('Importing data cannot be undone,\n'
              'are you sure you want to continue'),
            QtGui.QMessageBox.Cancel,
            QtGui.QMessageBox.Ok,
        )
        if answer == QtGui.QMessageBox.Ok:
            return True
        return False

    def isComplete(self):
        return self._complete


class FinalPage(ProgressPage):
    """FinalPage is the final page in the import process"""

    title = ugettext_lazy('Import Progress')
    sub_title = ugettext_lazy('Please wait while data is being imported.')

    def __init__(self, parent=None, model=None, admin=None):
        """
        :model: the source model from which to import data
        :admin: the admin class of the target data
        """
        super(FinalPage, self).__init__(parent)
        self.model = model
        self.admin = admin
        icon = 'tango/32x32/mimetypes/x-office-spreadsheet.png'
        self.setPixmap(QtGui.QWizard.LogoPixmap, Pixmap(icon).getQPixmap())
        self.setButtonText(QtGui.QWizard.FinishButton, _('Close'))
        self.progressbar = QtGui.QProgressBar()

    def run(self):
        collection = self.model.get_collection()
        self.update_maximum_signal.emit( len(collection) )
        for i,row in enumerate(collection):
            new_entity_instance = self.admin.entity()
            for field_name, attributes in self.model.get_admin().get_columns():
                try:
                    from_string = attributes['from_string']
                except KeyError:
                    logger.warn( 'field %s has no from_string field attribute, dont know how to import it properly'%attributes['original_field'] )
                    from_string = lambda _a:None
                setattr(
                    new_entity_instance,
                    attributes['original_field'],
                    from_string(getattr(row, field_name))
                )
            self.admin.add(new_entity_instance)
            self.admin.flush(new_entity_instance)
            self.update_progress_signal.emit(
                i, _('Row %i of %i imported') % (i+1, len(collection))
            )


class DataPreviewCollectionProxy(CollectionProxy):
    header_icon = None


class ImportWizard(QtGui.QWizard):
    """ImportWizard provides a two-step wizard for importing data as objects
    into Camelot.  To create a custom wizard, subclass this ImportWizard and
    overwrite its class attributes.

    To import a different file format, you probably need a custom
    collection_getter for this file type.
    """

    select_file_page = SelectFilePage
    data_preview_page = DataPreviewPage
    final_page = FinalPage
    collection_getter = CsvCollectionGetter
    window_title = _('Import CSV data')
    rowdata_admin_decorator = RowDataAdminDecorator

    def __init__(self, parent=None, admin=None):
        """:param admin: camelot model admin of the destination data"""
        super(ImportWizard, self).__init__(parent)
        assert admin
        #
        # Set the size of the wizard to 2/3rd of the screen, since we want to
        # get some work done here, the user needs to verify and possibly
        # correct its data
        #
        desktop = QtCore.QCoreApplication.instance().desktop()
        self.setMinimumSize(desktop.width()*2/3, desktop.height()*2/3)

        row_data_admin = self.rowdata_admin_decorator(admin)
        model = DataPreviewCollectionProxy(
            row_data_admin,
            lambda:[],
            row_data_admin.get_columns
        )
        self.setWindowTitle(self.window_title)
        self.add_pages(model, admin)
        self.setOption(QtGui.QWizard.NoCancelButton)

    def add_pages(self, model, admin):
        """
        Add all pages to the import wizard, reimplement this method to add
        custom pages to the wizard.  This method is called in the __init__
        method, to add all pages to the wizard.

        :param model: the CollectionProxy that will be used to display the to
        be imported data
        :param admin: the admin of the destination data
        """
        self.addPage(SelectFilePage(parent=self))
        self.addPage(
            DataPreviewPage(
                parent=self,
                model=model,
                collection_getter=self.collection_getter
            )
        )
        self.addPage(FinalPage(parent=self, model=model, admin=admin))






#  ============================================================================
#
#  Copyright (C) 2007-2011 Conceptive Engineering bvba. All rights reserved.
#  www.conceptive.be / project-camelot@conceptive.be
#
#  This file is part of the Camelot Library.
#
#  This file may be used under the terms of the GNU General Public
#  License version 2.0 as published by the Free Software Foundation
#  and appearing in the file license.txt included in the packaging of
#  this file.  Please review this information to ensure GNU
#  General Public Licensing requirements will be met.
#
#  If you are unsure which license is appropriate for your use, please
#  visit www.python-camelot.com or contact project-camelot@conceptive.be
#
#  This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
#  WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
#
#  For use of this library in commercial applications, please contact
#  project-camelot@conceptive.be
#
#  ============================================================================
from PyQt4 import QtCore

from wideeditor import WideEditor
from customeditor import AbstractCustomEditor, QtGui

class TextEditEditor(QtGui.QTextEdit, AbstractCustomEditor, WideEditor):

    editingFinished = QtCore.pyqtSignal()
 
    def __init__(self, 
                 parent, 
                 length=20, 
                 editable=True, 
                 field_name = 'text',
                 **kwargs):
        QtGui.QTextEdit.__init__(self, parent)
        self.setObjectName( field_name )
        AbstractCustomEditor.__init__(self)
        self.setReadOnly(not editable)

    def set_value(self, value):
        value = AbstractCustomEditor.set_value(self, value)
        #if value:
        #    self.setText(unicode(value))
        #else:
        #    self.setText('')
        self.setText(unicode(value))
        return value

    def get_value(self):
        val = AbstractCustomEditor.get_value(self)
        if val is not None: # we need to distinguish between None
            return val      # and other falsy values
        return unicode(self.toPlainText())


    def set_enabled(self, editable=True):
        self.setEnabled(editable)





import logging

FORMAT = '[%(levelname)-7s] [%(name)-35s] - %(message)s'
logging.basicConfig(level=logging.INFO, format=FORMAT)

logger = logging.getLogger('videostore.main')

try:
    import matplotlib
except:
    logger.error('Charts will not work because of missing matplotlib')

from camelot.core.conf import settings, SimpleSettings

class ExampleSettings( SimpleSettings ):
    """Special settings class for the example application, this is done to
    'survive' various packaging regimes, such as windows, debian, ...
    """
    
    @staticmethod
    def setup_model():
        import camelot.model
        import camelot_example.model
        from elixir import setup_all
        setup_all(create_tables=True)
        from camelot.model.authentication import updateLastLogin
        updateLastLogin()
        # 
        # Load sample data with the fixure mechanism
        #
        from camelot_example.fixtures import load_movie_fixtures
        load_movie_fixtures()
        from camelot.core.sql import update_database_from_model
        #update_database_from_model()
        #
        # setup the views
        #
        from camelot_example.view import setup_views
        setup_views()

settings.append( ExampleSettings('camelot', 'videostore') )

def main():
    from camelot.view.main import main
    from camelot_example.application_admin import MyApplicationAdmin
    main(MyApplicationAdmin())
    
if __name__ == '__main__':
    main()



#  ============================================================================
#
#  Copyright (C) 2007-2011 Conceptive Engineering bvba. All rights reserved.
#  www.conceptive.be / project-camelot@conceptive.be
#
#  This file is part of the Camelot Library.
#
#  This file may be used under the terms of the GNU General Public
#  License version 2.0 as published by the Free Software Foundation
#  and appearing in the file license.txt included in the packaging of
#  this file.  Please review this information to ensure GNU
#  General Public Licensing requirements will be met.
#
#  If you are unsure which license is appropriate for your use, please
#  visit www.python-camelot.com or contact project-camelot@conceptive.be
#
#  This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
#  WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
#
#  For use of this library in commercial applications, please contact
#  project-camelot@conceptive.be
#
#  ============================================================================

"""custom tree and tree-items widgets"""

import logging
logger = logging.getLogger('camelot.view.controls.modeltree')

from PyQt4 import QtGui
from PyQt4.QtCore import Qt

from camelot.core.utils import ugettext as _

class ModelItem(QtGui.QTreeWidgetItem):
    """Custom tree item widget"""

    def __init__(self, parent, columns_names, section_item):
        logger.debug('creating new modelitem')
        super(ModelItem, self).__init__(parent, columns_names)
        
        self.textColumn = 0
        self.iconColumn = 1
        self.section_item = section_item

        for column in (self.textColumn, self.iconColumn):
            self.setToolTip(column, _('Right click to open in New Tab'))

    def _underline(self, enable=False):
        font = self.font(self.textColumn)
        font.setUnderline(enable)
        self.setFont(self.textColumn, font)

    def set_icon(self, icon):
        self.setIcon(self.iconColumn, icon)
        
class ModelTree(QtGui.QTreeWidget):
    """Custom tree widget"""

    def __init__(self, header_labels=[''], parent=None):
        logger.debug('creating new modeltree')
        super(ModelTree, self).__init__(parent)
        # we don't select entire rows
        self.setSelectionBehavior(QtGui.QAbstractItemView.SelectItems)
        
        # we track mouse movement when no button is pressed
        self.setMouseTracking(True)
        self.header_labels = header_labels
        
        self.setColumnCount(2)
        self.setColumnWidth(0, 160)
        self.setColumnWidth(1, 18)
        self.setHorizontalScrollBarPolicy( Qt.ScrollBarAlwaysOff )
        self.setSelectionBehavior( self.SelectRows )
        
        self.clear_model_items()
        self.clear_section_items()
        self.fix_header_labels()

    def resizeEvent(self, event):
        self.setColumnWidth(0, self.width()  - 30 )
        
    def fix_header_labels(self):
        self.setHeaderHidden(True)

    def clear_section_items(self):
        self.section_items = []
        
    def clear_model_items(self):
        self.modelitems = []

    def mousePressEvent(self, event):
        """Custom context menu"""
        if event.button() == Qt.RightButton:
            self.customContextMenuRequested.emit( event.pos() )
            event.accept()
        else:
            QtGui.QTreeWidget.mousePressEvent(self, event)

    def leaveEvent(self, event):
        if not self.modelitems: return
        for item in self.modelitems: item._underline(False)

    def mouseMoveEvent(self, event):
        if not self.modelitems: return
        for item in self.modelitems: item._underline(False)

        item = self.itemAt(self.mapFromGlobal(self.cursor().pos()))
        if item:
            item._underline(True)

